#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Command line script to compute Pgens of CDR3 sequences.

    Copyright (C) 2018 Zachary Sethna

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

This script will compute  the generation probabilities (Pgens) of sequences, as
defined by a specified generative model. The sequences must be TRIMMED TO ONLY
THE CDR3 region as defined by the V and J anchor files (default is to INCLUDE
the conserved residues of the C in the V region and the F/W in the J region).

Each sequence will be determined to be one of:
1) 'Amino acid' sequence including any ambiguous amino acid symbols (see
Options for alphabet_file to specify custom symbols)
2) Nucleotide sequence (in-frame)
3) Regular expression template for 'amino acid sequences'

There are multiple default generative models that ship with OLGA and can be
specified with a flag:
--humanTRA (Human T cell alpha chain VJ model)
--humanTRB (Human T cell beta chain VDJ model)
--mouseTRB (Mouse T cell beta chain VDJ model)
--humanIGH (Human B cell heavy chain VDJ model)
--humanIGK (Human B cell light kappa chain VJ model)
--humanIGL (Human B cell light lambda chain VJ model)
--mouseIGH (Mouse B cell heavy chain VDJ model)
--mouseIGK (Mouse B cell light kappa chain VJ model)
--mouseIGL (Mouse B cell light lambda chain VJ model)

A custom model can also be specified (see below for details).

This script can read in sequences and output pgens in one of three modes:

1) Pass CDR3 sequences as arguments, output is printed to stdout.
2) Read in CDR3 sequences from a file, output is printed to stdout.
3) Read in CDR3 sequences from a file, output to a file, dynamic display with
time updates printed to stdout.


Mode 1):

This mode is provided as a way to quickly compute pgen of just a single or
couple of sequences. It is not advisable to use this mode to set up a loop
over a lot of sequences as each call of the script demands the overhead of
processing a model. To compute the pgens of many sequences, it is suggested to
read the sequences in from a file and use either mode 2 or 3.

It is also possible to condition the Pgen computation on V and J identity by
specifying the V or J usages as a mask. However, note that these V/J masks will
be applied to ALL of the sequences provided as arguments. Read Options on
v_mask and j_mask for more info.

--------------------------------------------------------------------------------
Example calls:

$ olga-compute_pgen --humanTRB CASSTGQANYGYTF
Pgen of the amino acid sequence CASSTGQANYGYTF: 5.26507446955e-08

$ olga-compute_pgen --humanTRB TGTGCCAGCAGTGACGCACAGGGGCGTAATCGTGGGACTGAAGCTTTCTTT
Pgen of the nucleotide sequence TGTGCCAGCAGTGACGCACAGGGGCGTAATCGTGGGACTGAAGCTTTCTTT: 1.31873701121e-17
Pgen of the amino acid sequence nt2aa(TGTGCCAGCAGTGACGCACAGGGGCGTAATCGTGGGACTGAAGCTTTCTTT) = CASSDAQGRNRGTEAFF: 4.70599549953e-13

#For a regular expression sequence, backslashes may be needed to specify the
#characters {} if seq read in as args
$ olga-compute_pgen --humanTRB CASSTGX\{1,5\}QAN[YA]GYTF
Pgen of the regular expression sequence CASSTGX{1,5}QAN[YA]GYTF: 7.588241802e-08

$ olga-compute_pgen --humanTRB CASSTGQANYGYTF CASSTGX\{1,5\}QAN[YA]GYTF TGTGCCAGCAGTGACGCACAGGGGCGTAATCGTGGGACTGAAGCTTTCTTT
Pgen of the amino acid sequence CASSTGQANYGYTF: 5.26507446955e-08
Pgen of the regular expression sequence CASSTGX{1,5}QAN[YA]GYTF: 7.588241802e-08
Pgen of the nucleotide sequence TGTGCCAGCAGTGACGCACAGGGGCGTAATCGTGGGACTGAAGCTTTCTTT: 1.31873701121e-17
Pgen of the amino acid sequence nt2aa(TGTGCCAGCAGTGACGCACAGGGGCGTAATCGTGGGACTGAAGCTTTCTTT) = CASSDAQGRNRGTEAFF: 4.70599549953e-13

$ olga-compute_pgen --humanTRB CASSTGQANYGYTF --v_mask TRBV14 --j_mask TRBJ1-2
Pgen of the amino acid sequence CASSTGQANYGYTF: 5.5513032863e-10

--------------------------------------------------------------------------------

Modes 2/3):

These read in sequences from a file. The script has only minimal file parsing
built in, so reading in sequences from a file requires the file to be structured
with delimiter spaced values (i.e. the  data is organized in columns separated
by delimiter like a .tsv or .csv file). Read Options on delimiter for more info.

It is not recommended to read in regular expression sequences from a file. These
sequences require enumerating out the amino acid sequences which correspond to
them and computing pgen for each of them individually -- this can require a
large time cost. Instead consider defining a custom 'amino acid' alphabet to
define the symbols used in the regular expressions if possible. Furthermore,
BE CAREFUL if reading in from a .csv file -- if commas are used in a regex
sequence and comma is used as the delimiter of the .csv file, the sequence will
not be read in properly.

If nucleotide sequences are to be read in it is possible to specify if the
output should be the nucleotide sequence Pgen and/or the translated amino acid
sequence Pgen (the default is to compute and output both). See Options.

To read in sequences, the index of column of CDR3 sequences is needed. The
default is to assume that the sequences to be read in are in the first column
(index 0), meaning that a text file with only a sequence on each line will be
read in okay by default. Read Options on seq_in for more info.

It is also possible to condition the Pgen computation on V and J identity by
specifying what index the column that V and J masks are stored for each line.

Mode 2 does not have a specified output file and so will print the sequences
and their pgens to stdout.

Mode 3 does have a specified output file. By default in this mode there is a
running display of the last few sequences/pgens written to the output file as
well as time elapsed, current rate of computation, and estimated time remaining.
This display can be disabled (see Options).

As it is rare for datasets to be >> 1e4, parallelization is not built in.
However, there are options to skip N lines of the file and to load at most M
sequences so, if wanted, one could build a parallelized wrapper around this
script (though it would be recommended to instead just import the modules and
build from there).

--------------------------------------------------------------------------------
Example calls (assumes a file example_seqs.tsv with the line structure
ntseq   aaseq   V_mask  J_mask):

#Reads in the ntseqs, prints the ntseq, aaseq and their pgens to stdout
$ olga-compute_pgen -i example_seqs.tsv --humanTRB

#Reads in the ntseqs, writes the ntseq, aaseq and their pgens to example_pgens.tsv
$ olga-compute_pgen -i example_seqs.tsv --humanTRB -o example_pgens.tsv

#Specifies the V/J mask indices
$ olga-compute_pgen -i example_seqs.tsv --humanTRB -o example_pgens.tsv --v_in 2 --j_in 3

#Reads in the aaseq column
$ olga-compute_pgen -i example_seqs.tsv --humanTRB -o example_pgens.tsv --seq_in 1

#Only runs the first 100 sequences
$ olga-compute_pgen -i example_seqs.tsv --humanTRB -o example_pgens.tsv --seq_in 1 -m 100
--------------------------------------------------------------------------------

To specify a custom model folder use:
--set_custom_model_VJ (generative model of VJ recombination, e.g. T alpha chain)
--set_custom_model_VDJ (generative model of VDJ recombination, e.g. T beta chain)

Note, if specifying a custom model folder for either a VJ recombination model
(e.g. an alpha or light chain model) or a VDJ recombination model
(e.g. a beta or heavy chain model), the folder must contain the following files
with the exact naming convention:

model_params.txt (IGoR inference param file)
model_marginals.txt (IGoR inference marginal file)
V_gene_CDR3_anchors.csv (V residue anchor and functionality file)
J_gene_CDR3_anchors.csv (J residue anchor and functionality file)

--------------------------------------------------------------------------------
Options:
  -h, --help            show this help message and exit
  --humanTRA, --human_T_alpha
                        use default human TRA model (T cell alpha chain)
  --humanTRB, --human_T_beta
                        use default human TRB model (T cell beta chain)
  --mouseTRB, --mouse_T_beta
                        use default mouse TRB model (T cell beta chain)
  --mouseTRA, --mouse_T_alpha
                        use default mouse TRA model (T cell alpha chain)
  --humanIGH, --human_B_heavy
                        use default human IGH model (B cell heavy chain)
  --humanIGK
                        use default human IGK model (B cell light kappa chain)
  --humanIGL, --human_B_lambda
                        use default human IGL model (B cell light lambda chain)
  --mouseIGH, --mouse_B_heavy
                        use default mouse IGH model (B cell heavy chain)
  --mouseIGK, --mouse_B_kappa
                        use default mouse IGK model (B cell light kappa chain)
  --mouseIGL, --mouse_B_lambda
                        use default mouse IGL model (B cell light lambda chain)
  --set_custom_model_VDJ=PATH/TO/FOLDER/
                        specify PATH/TO/FOLDER/ for a custom VDJ generative
                        model
  --set_custom_model_VJ=PATH/TO/FOLDER/
                        specify PATH/TO/FOLDER/ for a custom VJ generative
                        model
  -i PATH/TO/FILE, --infile=PATH/TO/FILE
                        read in CDR3 sequences (and optionally V/J masks) from
                        PATH/TO/FILE
  -o PATH/TO/FILE, --outfile=PATH/TO/FILE
                        write CDR3 sequences and pgens to PATH/TO/FILE
  --seq_in=INDEX, --seq_index=INDEX
                        specifies sequences to be read in are in column INDEX.
                        Default is index 0 (the first column).
  --v_in=INDEX, --v_mask_index=INDEX
                        specifies V_masks are found in column INDEX in the
                        input file. Default is no V mask.
  --j_in=INDEX, --j_mask_index=INDEX
                        specifies J_masks are found in column INDEX in the
                        input file. Default is no J mask.
  --v_mask=V_MASK       specify V usage to condition Pgen on for seqs read in
                        as arguments.
  --j_mask=J_MASK       specify J usage to condition Pgen on for seqs read in
                        as arguments.
  -m N, --max_number_of_seqs=N
                        compute Pgens for at most N sequences.
  --lines_to_skip=N     skip the first N lines of the file. Default is 0.
  -a PATH/TO/FILE, --alphabet_filename=PATH/TO/FILE
                        specify PATH/TO/FILE defining a custom 'amino acid'
                        alphabet. Default is no custom alphabet.
  --seq_type_out=SEQ_TYPE
                        if read in sequences are ntseqs, declare what type of
                        sequence to compute pgen for. Default is all. Choices:
                        'all', 'ntseq', 'nucleotide', 'aaseq', 'amino_acid'
  --skip_off, --skip_empty_off
                        stop skipping empty or blank sequences/lines (if for
                        example you want to keep line index fidelity between
                        the infile and outfile).
  --display_off         turn the sequence display off (only applies in safe
                        mode). Default is on.
  --num_lines_for_display=N
                        N lines of the output file are displayed when sequence
                        display is on. Also used to determine the number of
                        sequences to average over for speed and time
                        estimates.
  --time_updates_off    turn time updates off (only applies when sequence
                        display is disabled).
  --seqs_per_time_update=N
                        specify the number of sequences between time updates.
                        Default is 1e5.
  -d DELIMITER, --delimiter=DELIMITER
                        declare infile delimiter. Default is tab for .tsv
                        input files, comma for .csv files, and any whitespace
                        for all others. Choices: 'tab', 'space', ',', ';', ':'
  --raw_delimiter=DELIMITER
                        declare infile delimiter as a raw string.
  --delimiter_out=DELIMITER_OUT
                        declare outfile delimiter. Default is tab for .tsv
                        output files, comma for .csv files, and the infile
                        delimiter for all others. Choices: 'tab', 'space',
                        ',', ';', ':'
  --raw_delimiter_out=DELIMITER_OUT
                        declare for the delimiter outfile as a raw string.
  --gene_mask_delimiter=GENE_MASK_DELIMITER
                        declare gene mask delimiter. Default comma unless
                        infile delimiter is comma, then default is a
                        semicolon. Choices: 'tab', 'space', ',', ';', ':'
  --raw_gene_mask_delimiter=GENE_MASK_DELIMITER
                        declare delimiter of gene masks as a raw string.
  --comment_delimiter=COMMENT_DELIMITER
                        character or string to indicate comment or header
                        lines to skip.
  --fast_pgen           Use the numba implementation to calculate Pgen,
                        which is much faster.
--------------------------------------------------------------------------------


@author: zacharysethna
"""

#Function assumes that it is in the same directory that the folder app/ is
#in (which should contain all the modules imported).
from __future__ import print_function, division
import os
import sys
import time
import subprocess
reqs = subprocess.check_output([sys.executable, '-m', 'pip', 'freeze'])
installed_packages = [r.decode().split('==')[0] for r in reqs.split()]
if 'olga' not in installed_packages: sys.path.insert(0, os.path.split(os.path.dirname(os.path.abspath(__file__)))[0])
import olga.load_model as load_model
import olga.generation_probability as generation_probability
from olga.utils import nt2aa, determine_seq_type, gene_to_num_str
from olga.performance.fast_pgen import FastPgen
#
#import load_model
#import generation_probability
#from utils import nt2aa, determine_seq_type, gene_to_num_str

from optparse import OptionParser

#Set input = raw_input for python 2
try:
    import __builtin__
    input = getattr(__builtin__, 'raw_input')
except (ImportError, AttributeError):
    pass


#Need to determine what mode to run in.
#1) Sequences as args
#2) 'Safe mode', i.e. reading from file, writing to file
#3) not 'safe mode'. i.e. reading from file, printing to stdout


def main():
    """Compute Pgens from a file and output to another file."""

    parser = OptionParser(conflict_handler="resolve")
    parser.add_option('--humanTRA', '--human_T_alpha', action='store_true', dest='humanTRA', default=False, help='use default human TRA model (T cell alpha chain)')
    parser.add_option('--humanTRB', '--human_T_beta', action='store_true', dest='humanTRB', default=False, help='use default human TRB model (T cell beta chain)')
    parser.add_option('--mouseTRB', '--mouse_T_beta', action='store_true', dest='mouseTRB', default=False, help='use default mouse TRB model (T cell beta chain)')
    parser.add_option("--mouseTRA", '--mouse_T_alpha', action='store_true', dest='mouseTRA', default=False, help='use default mouse TRA model (T cell alpha chain)')
    parser.add_option('--humanIGH', '--human_B_heavy', action='store_true', dest='humanIGH', default=False, help='use default human IGH model (B cell heavy chain)')
    parser.add_option('--humanIGL', '--human_B_lambda', action='store_true', dest='humanIGL', default=False, help='use default human IGL model (B cell light lambda chain)')
    parser.add_option('--humanIGK', '--human_B_kappa', action='store_true', dest='humanIGK', default=False, help='use default human IGK model (B cell light kappa chain)')
    parser.add_option('--mouseIGH', '--mouse_B_heavy', action='store_true', dest='mouseIGH', default=False, help='use default mouse IGH model (B cell heavy chain)')
    parser.add_option('--mouseIGK', '--mouse_B_kappa', action='store_true', dest='mouseIGK', default=False, help='use default mouse IGK model (B cell light kappa chain)')
    parser.add_option('--mouseIGL', '--mouse_B_lambda', action='store_true', dest='mouseIGL', default=False, help='use default mouse IGL model (B cell light lambda chain)')
    parser.add_option('--set_custom_model_VDJ', dest='vdj_model_folder', metavar='PATH/TO/FOLDER/', help='specify PATH/TO/FOLDER/ for a custom VDJ generative model')
    parser.add_option('--set_custom_model_VJ', dest='vj_model_folder', metavar='PATH/TO/FOLDER/', help='specify PATH/TO/FOLDER/ for a custom VJ generative model')

    parser.add_option('-i', '--infile', dest = 'infile_name',metavar='PATH/TO/FILE', help='read in CDR3 sequences (and optionally V/J masks) from PATH/TO/FILE')
    parser.add_option('-o', '--outfile', dest = 'outfile_name', metavar='PATH/TO/FILE', help='write CDR3 sequences and pgens to PATH/TO/FILE')
    parser.add_option('--seq_in', '--seq_index', type='int', metavar='INDEX', dest='seq_in_index', default = 0, help='specifies sequences to be read in are in column INDEX. Default is index 0 (the first column).')

    parser.add_option('--v_in', '--v_mask_index', type='int', metavar='INDEX', dest='V_mask_index', help='specifies V_masks are found in column INDEX in the input file. Default is no V mask.')
    parser.add_option('--j_in', '--j_mask_index', type='int', metavar='INDEX', dest='J_mask_index', help='specifies J_masks are found in column INDEX in the input file. Default is no J mask.')

    parser.add_option('--v_mask', type='string', dest='V_mask', help='specify V usage to condition Pgen on for seqs read in as arguments.')
    parser.add_option('--j_mask', type='string', dest='J_mask', help='specify J usage to condition Pgen on for seqs read in as arguments.')

    parser.add_option('-m', '--max_number_of_seqs', type='int',metavar='N', dest='max_number_of_seqs', help='compute Pgens for at most N sequences.')
    parser.add_option('--lines_to_skip', type='int',metavar='N', dest='lines_to_skip', default = 0, help='skip the first N lines of the file. Default is 0.')
    parser.add_option('-a', '--alphabet_filename', dest='alphabet_filename', metavar='PATH/TO/FILE', help="specify PATH/TO/FILE defining a custom 'amino acid' alphabet. Default is no custom alphabet.")
    parser.add_option('--seq_type_out', type='choice',metavar='SEQ_TYPE', dest='seq_type_out',  choices=['all', 'ntseq', 'nucleotide', 'aaseq', 'amino_acid'], help="if read in sequences are ntseqs, declare what type of sequence to compute pgen for. Default is all. Choices: 'all', 'ntseq', 'nucleotide', 'aaseq', 'amino_acid'")
    parser.add_option('--skip_off','--skip_empty_off', action='store_true', dest = 'skip_empty', default=True, help='stop skipping empty or blank sequences/lines (if for example you want to keep line index fidelity between the infile and outfile).')

    parser.add_option('--display_off', action='store_false', dest='display_seqs', default=True, help='turn the sequence display off (only applies in write-to-file mode). Default is on.')
    parser.add_option('--num_lines_for_display', type='int', metavar='N', default = 50, dest='num_lines_for_display', help='N lines of the output file are displayed when sequence display is on. Also used to determine the number of sequences to average over for speed and time estimates.')
    parser.add_option('--time_updates_off', action='store_false', dest='time_updates', default=True, help='turn time updates off (only applies when sequence display is disabled).')
    parser.add_option('--seqs_per_time_update', type='float', metavar='N', default = 100, dest='seqs_per_time_update', help='specify the number of sequences between time updates. Default is 1e5.')

    parser.add_option('-d', '--delimiter', type='choice', dest='delimiter',  choices=['tab', 'space', ',', ';', ':'], help="declare infile delimiter. Default is tab for .tsv input files, comma for .csv files, and any whitespace for all others. Choices: 'tab', 'space', ',', ';', ':'")
    parser.add_option('--raw_delimiter', type='str', dest='delimiter', help="declare infile delimiter as a raw string.")
    parser.add_option('--delimiter_out', type='choice', dest='delimiter_out',  choices=['tab', 'space', ',', ';', ':'], help="declare outfile delimiter. Default is tab for .tsv output files, comma for .csv files, and the infile delimiter for all others. Choices: 'tab', 'space', ',', ';', ':'")
    parser.add_option('--raw_delimiter_out', type='str', dest='delimiter_out', help="declare for the delimiter outfile as a raw string.")
    parser.add_option('--gene_mask_delimiter', type='choice', dest='gene_mask_delimiter',  choices=['tab', 'space', ',', ';', ':'], help="declare gene mask delimiter. Default comma unless infile delimiter is comma, then default is a semicolon. Choices: 'tab', 'space', ',', ';', ':'")
    parser.add_option('--raw_gene_mask_delimiter', type='str', dest='gene_mask_delimiter', help="declare delimiter of gene masks as a raw string.")
    parser.add_option('--comment_delimiter', type='str', dest='comment_delimiter', help="character or string to indicate comment or header lines to skip.")
    parser.add_option('--skip_fast_pgen', action='store_true', dest='skip_fast_pgen', default=False, help='Skip the numba implementation to calculate Pgen, which is much faster.')

    (options, args) = parser.parse_args()

    #Check that the model is specified properly
    main_folder = os.path.dirname(__file__)

    default_models = {}
    default_models['humanTRA'] = [os.path.join(main_folder, 'default_models', 'human_T_alpha'),  'VJ']
    default_models['humanTRB'] = [os.path.join(main_folder, 'default_models', 'human_T_beta'), 'VDJ']
    default_models['mouseTRB'] = [os.path.join(main_folder, 'default_models', 'mouse_T_beta'), 'VDJ']
    default_models['mouseTRA'] = [os.path.join(main_folder, 'default_models', 'mouse_T_alpha'), 'VJ']
    default_models['humanIGH'] = [os.path.join(main_folder, 'default_models', 'human_B_heavy'), 'VDJ']
    default_models['humanIGK'] = [os.path.join(main_folder, 'default_models', 'human_B_kappa'), 'VJ']
    default_models['humanIGL'] = [os.path.join(main_folder, 'default_models', 'human_B_lambda'),  'VJ']
    default_models['mouseIGH'] = [os.path.join(main_folder, 'default_models', 'mouse_B_heavy'), 'VDJ']
    default_models['mouseIGK'] = [os.path.join(main_folder, 'default_models', 'mouse_B_kappa'), 'VJ']
    default_models['mouseIGL'] = [os.path.join(main_folder, 'default_models', 'mouse_B_lambda'), 'VJ']

    num_models_specified = sum([1 for x in list(default_models.keys()) + ['vj_model_folder', 'vdj_model_folder'] if getattr(options, x)])

    if num_models_specified == 1: #exactly one model specified
        try:
            d_model = [x for x in default_models.keys() if getattr(options, x)][0]
            model_folder = default_models[d_model][0]
            recomb_type = default_models[d_model][1]
        except IndexError:
            if options.vdj_model_folder: #custom VDJ model specified
                model_folder = options.vdj_model_folder
                recomb_type = 'VDJ'
            elif options.vj_model_folder: #custom VJ model specified
                model_folder = options.vj_model_folder
                recomb_type = 'VJ'
    elif num_models_specified == 0:
        print('Need to indicate generative model.')
        print('Exiting...')
        return -1
    elif num_models_specified > 1:
        print('Only specify one model')
        print('Exiting...')
        return -1

    #Check that all model and genomic files exist in the indicated model folder
    if not os.path.isdir(model_folder):
        print('Check pathing... cannot find the model folder: ' + model_folder)
        print('Exiting...')
        return -1

    params_file_name = os.path.join(model_folder,'model_params.txt')
    marginals_file_name = os.path.join(model_folder,'model_marginals.txt')
    V_anchor_pos_file = os.path.join(model_folder,'V_gene_CDR3_anchors.csv')
    J_anchor_pos_file = os.path.join(model_folder,'J_gene_CDR3_anchors.csv')

    for x in [params_file_name, marginals_file_name, V_anchor_pos_file, J_anchor_pos_file]:
        if not os.path.isfile(x):
            print('Cannot find: ' + x)
            print('Please check the files (and naming conventions) in the model folder ' + model_folder)
            print('Exiting...')
            return -1

    alphabet_filename = options.alphabet_filename #used if a custom alphabet is to be specified
    if alphabet_filename is not None:
        if not os.path.isfile(alphabet_filename):
            print('Cannot find custom alphabet file: ' + alphabet_filename)
            print('Exiting...')
            return -1

    #Load up model based on recomb_type
    #VDJ recomb case --- used for TCRB and IGH
    if recomb_type == 'VDJ':
        genomic_data = load_model.GenomicDataVDJ()
        genomic_data.load_igor_genomic_data(params_file_name, V_anchor_pos_file, J_anchor_pos_file)
        generative_model = load_model.GenerativeModelVDJ()
        generative_model.load_and_process_igor_model(marginals_file_name)
        pgen_model = generation_probability.GenerationProbabilityVDJ(generative_model, genomic_data, alphabet_filename)
    #VJ recomb case --- used for TCRA and light chain
    elif recomb_type == 'VJ':
        genomic_data = load_model.GenomicDataVJ()
        genomic_data.load_igor_genomic_data(params_file_name, V_anchor_pos_file, J_anchor_pos_file)
        generative_model = load_model.GenerativeModelVJ()
        generative_model.load_and_process_igor_model(marginals_file_name)
        pgen_model = generation_probability.GenerationProbabilityVJ(generative_model, genomic_data, alphabet_filename)
    
    if not options.skip_fast_pgen:
        pgen_model = FastPgen(pgen_model)

    aa_alphabet = ''.join(pgen_model.codons_dict.keys())

    if options.infile_name is not None:
        infile_name = options.infile_name

        if not os.path.isfile(infile_name):
            print('Cannot find input file: ' + infile_name)
            print('Exiting...')
            return -1

    if options.outfile_name is not None:
        outfile_name = options.outfile_name
        if os.path.isfile(outfile_name):
            if not input(outfile_name + ' already exists. Overwrite (y/n)? ').strip().lower() in ['y', 'yes']:
                print('Exiting...')
                return -1

    #Parse delimiter
    delimiter = options.delimiter
    if delimiter is None: #Default case
        if options.infile_name is None:
            delimiter = '\t'
        elif infile_name.endswith('.tsv'): #parse TAB separated value file
            delimiter = '\t'
        elif infile_name.endswith('.csv'): #parse COMMA separated value file
            delimiter = ','
    else:
        try:
            delimiter = {'tab': '\t', 'space': ' ', ',': ',', ';': ';', ':': ':'}[delimiter]
        except KeyError:
            pass #Other string passed as the delimiter.

    #Parse delimiter_out
    delimiter_out = options.delimiter_out
    if delimiter_out is None: #Default case
        if delimiter is None:
            delimiter_out = '\t'
        else:
            delimiter_out = delimiter
        if options.outfile_name is None:
            pass
        elif outfile_name.endswith('.tsv'): #output TAB separated value file
            delimiter_out = '\t'
        elif outfile_name.endswith('.csv'): #output COMMA separated value file
            delimiter_out = ','
    else:
        try:
            delimiter_out = {'tab': '\t', 'space': ' ', ',': ',', ';': ';', ':': ':'}[delimiter_out]
        except KeyError:
            pass #Other string passed as the delimiter.

    #Parse gene_delimiter
    gene_mask_delimiter = options.gene_mask_delimiter
    if gene_mask_delimiter is None: #Default case
        gene_mask_delimiter = ','
        if delimiter == ',':
            gene_mask_delimiter = ';'
    else:
        try:
            gene_mask_delimiter = {'tab': '\t', 'space': ' ', ',': ',', ';': ';', ':': ':'}[gene_mask_delimiter]
        except KeyError:
            pass #Other string passed as the delimiter.


    #More options
    time_updates = options.time_updates
    display_seqs = options.display_seqs
    num_lines_for_display = options.num_lines_for_display
    seq_in_index = options.seq_in_index #where in the line the sequence is after line.split(delimiter)
    lines_to_skip = options.lines_to_skip #one method of skipping header
    comment_delimiter = options.comment_delimiter #another method of skipping header
    seqs_per_time_update = options.seqs_per_time_update
    max_number_of_seqs = options.max_number_of_seqs
    V_mask_index = options.V_mask_index #Default is not conditioning on V identity
    J_mask_index = options.J_mask_index #Default is not conditioning on J identity
    skip_empty = options.skip_empty

    seq_type_out = options.seq_type_out #type of pgens to be computed. Can be ntseq, aaseq, or both
    if seq_type_out is not None:
        seq_type_out = {'all': None, 'ntseq': 'ntseq', 'nucleotide': 'ntseq', 'aaseq': 'aaseq', 'amino_acid': 'aaseq'}[seq_type_out]

    if options.infile_name is None: #No infile specified -- args should be the input seqs
        print_warnings = True
        seqs = args
        seq_types = [determine_seq_type(seq, aa_alphabet) for seq in seqs]
        unrecognized_seqs = [seq for i, seq in enumerate(seqs) if seq_types[i] is None]
        if len(unrecognized_seqs) > 0 and print_warnings:
            print('The following sequences/arguments were not recognized: ' + ', '.join(unrecognized_seqs))
        seqs = [seq for i, seq in enumerate(seqs) if seq_types[i] is not None]
        seq_types = [seq_type for seq_type in seq_types if seq_type is not None]


        #Format V and J masks -- uniform for all argument input sequences
        try:
            V_mask = options.V_mask.split(',')
            unrecognized_v_genes = [v for v in V_mask if gene_to_num_str(v, 'V') not in pgen_model.V_mask_mapping.keys()]
            V_mask = [v for v in V_mask if gene_to_num_str(v, 'V') in pgen_model.V_mask_mapping.keys()]
            if len(unrecognized_v_genes) > 0:
                print('These V genes/alleles are not recognized: ' + ', '.join(unrecognized_v_genes))
            if len(V_mask) == 0:
                print('No recognized V genes/alleles in the provided V_mask. Continuing without conditioning on V usage.')
                V_mask = None
        except AttributeError:
            V_mask = options.V_mask #Default is None, i.e. not conditioning on V identity

        try:
            J_mask = options.J_mask.split(',')
            unrecognized_j_genes = [j for j in J_mask if gene_to_num_str(j, 'J') not in pgen_model.J_mask_mapping.keys()]
            J_mask = [j for j in J_mask if gene_to_num_str(j, 'J') in pgen_model.J_mask_mapping.keys()]
            if len(unrecognized_j_genes) > 0:
                print('These J genes/alleles are not recognized: ' + ', '.join(unrecognized_j_genes))
            if len(J_mask) == 0:
                print('No recognized J genes/alleles in the provided J_mask. Continuing without conditioning on J usage.')
                J_mask = None
        except AttributeError:
            J_mask = options.J_mask #Default is None, i.e. not conditioning on J identity

        print('')
        start_time = time.time()
        for seq, seq_type in zip(seqs, seq_types):
            if seq_type == 'aaseq':
                c_pgen = pgen_model.compute_aa_CDR3_pgen(seq, V_mask, J_mask, print_warnings)
                print('Pgen of the amino acid sequence ' + seq + ': ' + str(c_pgen))
                print('')
            elif seq_type == 'regex':
                c_pgen = pgen_model.compute_regex_CDR3_template_pgen(seq, V_mask, J_mask, print_warnings)
                print('Pgen of the regular expression sequence ' + seq + ': ' + str(c_pgen))
                print('')
            elif seq_type == 'ntseq':
                if seq_type_out is None or seq_type_out == 'ntseq':
                    c_pgen_nt = pgen_model.compute_nt_CDR3_pgen(seq, V_mask, J_mask, print_warnings)
                    print('Pgen of the nucleotide sequence ' + seq + ': ' + str(c_pgen_nt))
                if seq_type_out is None or seq_type_out == 'aaseq':
                    c_pgen_aa = pgen_model.compute_aa_CDR3_pgen(nt2aa(seq), V_mask, J_mask, print_warnings)
                    print('Pgen of the amino acid sequence nt2aa(' + seq + ') = ' + nt2aa(seq) + ': ' + str(c_pgen_aa))
                print('')

        c_time = time.time() - start_time
        if c_time > 86400: #more than a day
            c_time_str = '%d days, %d hours, %d minutes, and %.2f seconds.'%(int(c_time)//86400, (int(c_time)//3600)%24, (int(c_time)//60)%60, c_time%60)
        elif c_time > 3600: #more than an hr
            c_time_str = '%d hours, %d minutes, and %.2f seconds.'%((int(c_time)//3600)%24, (int(c_time)//60)%60, c_time%60)
        elif c_time > 60: #more than a min
            c_time_str = '%d minutes and %.2f seconds.'%((int(c_time)//60)%60, c_time%60)
        else:
            c_time_str = '%.2f seconds.'%(c_time)

        print('Completed pgen computation in: ' + c_time_str)

    else: #Read sequences in from file
        print_warnings = False #Most cases of reading in from file should have warnings disabled
        seqs = []
        seq_types = []
        V_usage_masks = []
        J_usage_masks = []

        infile = open(infile_name, 'r')

        for i, line in enumerate(infile):
            if comment_delimiter is not None: #Default case -- no comments/header delimiter
                if line.startswith(comment_delimiter): #allow comments
                    continue
            if i < lines_to_skip:
                continue

            if delimiter is None: #Default delimiter is any whitespace
                split_line = line.split()
            else:
                split_line = line.split(delimiter)

            #Find the seq
            try:
                seq = split_line[seq_in_index].strip()
                if len(seq.strip()) == 0:
                    if skip_empty:
                        continue
                    else:
                        seqs.append(seq) #keep the blank seq as a placeholder
                        seq_types.append('aaseq')
                else:
                    seqs.append(seq)
                    seq_types.append(determine_seq_type(seq, aa_alphabet))
            except IndexError: #no index match for seq
                if skip_empty and len(line.strip()) == 0:
                    continue
                print('seq_in_index is out of range')
                print('Exiting...')
                infile.close()
                return -1

            #Find and format V_usage_mask
            if V_mask_index is None:
                V_usage_masks.append(None) #default mask
            else:
                try:
                    V_usage_mask = split_line[V_mask_index].strip().split(gene_mask_delimiter)
                    #check that all V gene/allele names are recognized
                    if all([gene_to_num_str(v, 'V') in pgen_model.V_mask_mapping for v in V_usage_mask]):
                        V_usage_masks.append(V_usage_mask)
                    else:
                        print(str(V_usage_mask) + " is not a usable V_usage_mask composed exclusively of recognized V gene/allele names")
                        print('Unrecognized V gene/allele names: ' + ', '.join([v for v in V_usage_mask if gene_to_num_str(v, 'V') not in pgen_model.V_mask_mapping.keys()]))
                        print('Exiting...')
                        infile.close()
                        return -1
                except IndexError: #no index match for V_mask_index
                    print('V_mask_index is out of range')
                    print('Exiting...')
                    infile.close()
                    return -1

            #Find and format J_usage_mask
            if J_mask_index is None:
                J_usage_masks.append(None) #default mask
            else:
                try:
                    J_usage_mask = split_line[J_mask_index].strip().split(gene_mask_delimiter)
                    #check that all V gene/allele names are recognized
                    if all([gene_to_num_str(j, 'J') in pgen_model.J_mask_mapping for j in J_usage_mask]):
                        J_usage_masks.append(J_usage_mask)
                    else:
                        print(str(J_usage_mask) + " is not a usable J_usage_mask composed exclusively of recognized J gene/allele names")
                        print('Unrecognized J gene/allele names: ' + ', '.join([j for j in J_usage_mask if gene_to_num_str(j, 'J') not in pgen_model.J_mask_mapping.keys()]))
                        print('Exiting...')
                        infile.close()
                        return -1
                except IndexError: #no index match for J_mask_index
                    print('J_mask_index is out of range')
                    print('Exiting...')
                    infile.close()
                    return -1

            if max_number_of_seqs is not None:
                if len(seqs) >= max_number_of_seqs:
                    break


        unrecognized_seqs = [seq for i, seq in enumerate(seqs) if seq_types[i] is None]
        if len(unrecognized_seqs) > 0 and len(unrecognized_seqs) < len(seqs):
            if print_warnings or options.outfile_name is not None:
                print('Some strings read in were not parsed as sequences -- they will be omitted.')
                print('Examples of improperly read strings: ')
                for unrecognized_seq in unrecognized_seqs[:10]:
                    print(unrecognized_seq)
            seqs = [seq for i, seq in enumerate(seqs) if seq_types[i] is not None]
            V_usage_masks = [V_usage_mask for i, V_usage_mask in enumerate(V_usage_masks) if seq_types[i] is not None]
            J_usage_masks = [J_usage_mask for i, J_usage_mask in enumerate(J_usage_masks) if seq_types[i] is not None]
            seq_types = [seq_type for seq_type in seq_types if seq_type is not None]
        elif len(unrecognized_seqs) > 0 and len(unrecognized_seqs) == len(seqs):
            print('None of the read in strings were parsed as sequences. Check input file.')
            print('Examples of improperly read strings:')
            for unrecognized_seq in unrecognized_seqs[:10]:
                print(unrecognized_seq)
            print('Exiting...')
            return -1

        infile.close()


        if options.outfile_name is not None: #OUTFILE SPECIFIED, allow printed info/display

            print('Successfully read in and formatted ' + str(len(seqs)) + ' sequences and any V or J usages.')
            if display_seqs:
                sys.stdout.write('\r'+'Continuing to Pgen computation in 3... ')
                sys.stdout.flush()
                time.sleep(0.4)
                sys.stdout.write('\r'+'Continuing to Pgen computation in 2... ')
                sys.stdout.flush()
                time.sleep(0.4)
                sys.stdout.write('\r'+'Continuing to Pgen computation in 1... ')
                sys.stdout.flush()
                time.sleep(0.4)
            else:
                print('Continuing to Pgen computation.')
                print_warnings = True #Display is off, can print warnings

            if display_seqs:
                lines_for_display = []
                times_for_speed_calc = [time.time()]

            outfile = open(outfile_name, 'w')
            start_time = time.time()
            for i, seq in enumerate(seqs):
                if seq_types[i] == 'aaseq':
                    #Compute Pgen and print out
                    c_pgen_line = seq + delimiter_out + str(pgen_model.compute_aa_CDR3_pgen(seq, V_usage_masks[i], J_usage_masks[i], print_warnings))
                if seq_types[i] == 'regex':
                    #Compute Pgen and print out
                    c_pgen_line = seq + delimiter_out + str(pgen_model.compute_regex_CDR3_template_pgen(seq, V_usage_masks[i], J_usage_masks[i], print_warnings))
                elif seq_types[i] == 'ntseq':
                    ntseq = seq
                    if len(ntseq) % 3 == 0: #inframe sequence
                        aaseq = nt2aa(ntseq)
                        #Compute Pgen and print out based on recomb_type and seq_type_out
                        if seq_type_out is None:
                            c_pgen_line = ntseq + delimiter_out + str(pgen_model.compute_nt_CDR3_pgen(ntseq, V_usage_masks[i], J_usage_masks[i], print_warnings)) + delimiter_out + aaseq + delimiter_out +  str(pgen_model.compute_aa_CDR3_pgen(aaseq, V_usage_masks[i], J_usage_masks[i], print_warnings))
                        elif seq_type_out == 'ntseq':
                            c_pgen_line = ntseq + delimiter_out + str(pgen_model.compute_nt_CDR3_pgen(ntseq, V_usage_masks[i], J_usage_masks[i], print_warnings))
                        elif seq_type_out == 'aaseq':
                            c_pgen_line = aaseq + delimiter_out + str(pgen_model.compute_aa_CDR3_pgen(aaseq, V_usage_masks[i], J_usage_masks[i], print_warnings))
                    else: #out of frame sequence -- Pgens are 0 and use 'out_of_frame' for aaseq
                        if seq_type_out is None:
                            c_pgen_line = ntseq + delimiter_out + '0' + delimiter_out + 'out_of_frame' + delimiter_out + '0'
                        elif seq_type_out == 'ntseq':
                            c_pgen_line = ntseq + delimiter_out + '0'
                        elif seq_type_out == 'aaseq':
                            c_pgen_line = 'out_of_frame' + delimiter_out + '0'

                outfile.write(c_pgen_line + '\n')

                #Print time update
                if display_seqs:
                    cc_time = time.time()
                    c_time = cc_time - start_time
                    times_for_speed_calc = [cc_time] + times_for_speed_calc[:num_lines_for_display]
                    c_avg_speed = (len(times_for_speed_calc)-1)/(times_for_speed_calc[0] - times_for_speed_calc[-1])

                    #eta = ((len(seqs) - (i+1))/float(i+1))*c_time

                    eta = (len(seqs) - (i+1))/c_avg_speed

                    lines_for_display = [c_pgen_line] + lines_for_display[:num_lines_for_display]


                    c_time_str = '%s hours, %s minutes, and %s seconds.'%(repr(int(c_time)//3600).rjust(3), repr((int(c_time)//60)%60).rjust(2), repr(int(c_time)%60).rjust(2))
                    eta_str = '%s hours, %s minutes, and %s seconds.'%(repr(int(eta)//3600).rjust(3), repr((int(eta)//60)%60).rjust(2), repr(int(eta)%60).rjust(2))
                    time_str = 'Time to compute Pgen on %s seqs: %s \nEst. time for remaining %s seqs: %s'%(repr(i+1).rjust(9), c_time_str, repr(len(seqs) - (i + 1)).rjust(9), eta_str)
                    speed_str = 'Current Pgen computation speed: %s seqs/min'%(repr(round((len(times_for_speed_calc)-1)*60/(times_for_speed_calc[0] - times_for_speed_calc[-1]), 2)).rjust(8))
                    display_str = '\n'.join(lines_for_display[::-1]) + '\n' + '-'*80 + '\n' + time_str + '\n' + speed_str + '\n' + '-'*80
                    print('\033[2J' + display_str)
                elif (i+1)%seqs_per_time_update == 0 and time_updates:
                    c_time = time.time() - start_time
                    eta = ((len(seqs) - (i+1))/(i+1))*c_time
                    if c_time > 86400: #more than a day
                        c_time_str = '%d days, %d hours, %d minutes, and %.2f seconds.'%(int(c_time)//86400, (int(c_time)//3600)%24, (int(c_time)//60)%60, c_time%60)
                    elif c_time > 3600: #more than an hr
                        c_time_str = '%d hours, %d minutes, and %.2f seconds.'%((int(c_time)//3600)%24, (int(c_time)//60)%60, c_time%60)
                    elif c_time > 60: #more than a min
                        c_time_str = '%d minutes and %.2f seconds.'%((int(c_time)//60)%60, c_time%60)
                    else:
                        c_time_str = '%.2f seconds.'%(c_time)

                    if eta > 86400: #more than a day
                        eta_str = '%d days, %d hours, %d minutes, and %.2f seconds.'%(int(eta)//86400, (int(eta)//3600)%24, (int(eta)//60)%60, eta%60)
                    elif eta > 3600: #more than an hr
                        eta_str = '%d hours, %d minutes, and %.2f seconds.'%((int(eta)//3600)%24, (int(eta)//60)%60, eta%60)
                    elif eta > 60: #more than a min
                        eta_str = '%d minutes and %.2f seconds.'%((int(eta)//60)%60, eta%60)
                    else:
                        eta_str = '%.2f seconds.'%(eta)

                    print('Pgen computed for %d sequences in: %s Estimated time remaining: %s'%(i+1, c_time_str, eta_str))

            c_time = time.time() - start_time
            if c_time > 86400: #more than a day
                c_time_str = '%d days, %d hours, %d minutes, and %.2f seconds.'%(int(c_time)//86400, (int(c_time)//3600)%24, (int(c_time)//60)%60, c_time%60)
            elif c_time > 3600: #more than an hr
                c_time_str = '%d hours, %d minutes, and %.2f seconds.'%((int(c_time)//3600)%24, (int(c_time)//60)%60, c_time%60)
            elif c_time > 60: #more than a min
                c_time_str = '%d minutes and %.2f seconds.'%((int(c_time)//60)%60, c_time%60)
            else:
                c_time_str = '%.2f seconds.'%(c_time)
            print('Completed Pgen computation for %d sequences: in %s'%(len(seqs), c_time_str))

            outfile.close()

        else: #NO OUTFILE -- print directly to stdout
            start_time = time.time()
            for i, seq in enumerate(seqs):
                if seq_types[i] == 'aaseq':
                    #Compute Pgen and print out
                    c_pgen_line = seq + delimiter_out + str(pgen_model.compute_aa_CDR3_pgen(seq, V_usage_masks[i], J_usage_masks[i], print_warnings))
                if seq_types[i] == 'regex':
                    #Compute Pgen and print out
                    c_pgen_line = seq + delimiter_out + str(pgen_model.compute_regex_CDR3_template_pgen(seq, V_usage_masks[i], J_usage_masks[i], print_warnings))
                elif seq_types[i] == 'ntseq':
                    ntseq = seq
                    if len(ntseq) % 3 == 0: #inframe sequence
                        aaseq = nt2aa(ntseq)
                        #Compute Pgen and print out based on recomb_type and seq_type_out
                        if seq_type_out is None:
                            c_pgen_line = ntseq + delimiter_out + str(pgen_model.compute_nt_CDR3_pgen(ntseq, V_usage_masks[i], J_usage_masks[i], print_warnings)) + delimiter_out + aaseq + delimiter_out +  str(pgen_model.compute_aa_CDR3_pgen(aaseq, V_usage_masks[i], J_usage_masks[i], print_warnings))
                        elif seq_type_out == 'ntseq':
                            c_pgen_line = ntseq + delimiter_out + str(pgen_model.compute_nt_CDR3_pgen(ntseq, V_usage_masks[i], J_usage_masks[i], print_warnings))
                        elif seq_type_out == 'aaseq':
                            c_pgen_line = aaseq + delimiter_out + str(pgen_model.compute_aa_CDR3_pgen(aaseq, V_usage_masks[i], J_usage_masks[i], print_warnings))
                    else: #out of frame sequence -- Pgens are 0 and use 'out_of_frame' for aaseq
                        if seq_type_out is None:
                            c_pgen_line = ntseq + delimiter_out + '0' + delimiter_out + 'out_of_frame' + delimiter_out + '0'
                        elif seq_type_out == 'ntseq':
                            c_pgen_line = ntseq + delimiter_out + '0'
                        elif seq_type_out == 'aaseq':
                            c_pgen_line = 'out_of_frame' + delimiter_out + '0'

                print(c_pgen_line)

if __name__ == '__main__': main()



from typing import List, Union
from types import MethodType
import numpy as np

from olga.generation_probability import GenerationProbabilityVJ, GenerationProbabilityVDJ
from olga.performance.kernels import (
    compute_Pi_R_one_numba,
    compute_Pi_L_numba,
    compute_Pi_JinsDJ_given_D_numba,
    compute_Pi_V_insVJ_given_J_numba
)

try:
    import numba as nb
    NUMBA_INSTALLED = True
except ImportError:
    NUMBA_INSTALLED = False


class FastPgen:
    """ An wrapper around OLGA's `GenerationProbabilityVJ` or `GenerationProbabilityVDJ`
    object for calculating Pgen with a faster implementation of `compute_CDR3_pgen` thanks 
    to numba.

    Call any compute method such as `compute_aa_CDR3_pgen` or `compute_nt_CDR3_pgen`
    as usual, and the faster `compute_CDR3_pgen` will be used under the hood.
 
    Parameters
    ----------
    impl : Union[GenerationProbabilityVJ, GenerationProbabilityVDJ]
        An existing OLGA generation probability model.

    Examples
    --------
    >>> from olga import GenerationProbabilityVJ
    >>> base_model = GenerationProbabilityVJ(...)
    >>> fast_model = FastPgen(base_model)
    >>> p = pgen_model.compute_aa_CDR3_pgen('CAWSVAPDRGGYTF', 'TRBV30', 'TRBJ1-2')
    """

    def __init__(self, impl: Union[GenerationProbabilityVJ, GenerationProbabilityVDJ]) -> None:
        self._impl = impl

        if not NUMBA_INSTALLED:
            raise RuntimeError('FastPgen requires numba. ')

        self._init_common()
        if isinstance(self._impl, GenerationProbabilityVJ):
            self._init_vj()
        elif isinstance(self._impl, GenerationProbabilityVDJ):
            self._init_vdj()
        else:
            raise ValueError('Invalid model passed to FastPgen')

        # Monkey patch to override the default compute_CDR3_pgen
        def _proxy(impl_self, *args, **kwargs):
            return self.compute_CDR3_pgen(*args, **kwargs)
        impl.compute_CDR3_pgen = MethodType(_proxy, impl)
    

    def _init_common(self) -> None:
        self._nt_lut = np.full(256, -1, dtype=np.int8)
        for i, c in enumerate(b'ACGT'):
            self._nt_lut[c] = i

        self.i2aa = list(self.codons_dict.keys())
        self.aa2i = {aa:i for i, aa in enumerate(self.i2aa)}

        self._aa_lut = np.full(256, -1, dtype=np.int16)
        for aa, i in self.aa2i.items():
            self._aa_lut[ord(aa)] = i


    def _init_vj(self) -> None:
        # Need these to be pure numpy arrays to use in numba
        self.Svj_stack = np.stack([self.Svj[a] for a in self.i2aa])
        self.Dvj_stack = np.stack([self.Dvj[a] for a in self.i2aa])
        self.Tvj_stack = np.stack([self.Tvj[a] for a in self.i2aa])
        self.lDvj_stack = np.stack([self.lDvj[a] for a in self.i2aa])
        self.lTvj_stack = np.stack([self.lTvj[a] for a in self.i2aa])


    def _init_vdj(self) -> None:
        # Need these to be pure numpy arrays to use in numba
        self.Svd_stack = np.stack([self.Svd[a] for a in self.i2aa])
        self.Dvd_stack = np.stack([self.Dvd[a] for a in self.i2aa])
        self.Tvd_stack = np.stack([self.Tvd[a] for a in self.i2aa])
        self.lDvd_stack = np.stack([self.lDvd[a] for a in self.i2aa])
        self.lTvd_stack = np.stack([self.lTvd[a] for a in self.i2aa])

        self.Sdj_stack = np.stack([self.Sdj[a] for a in self.i2aa])
        self.Ddj_stack = np.stack([self.Ddj[a] for a in self.i2aa])
        self.Tdj_stack = np.stack([self.Tdj[a] for a in self.i2aa])
        self.rDdj_stack = np.stack([self.rDdj[a] for a in self.i2aa])
        self.rTdj_stack = np.stack([self.rTdj[a] for a in self.i2aa])

        # Preparing the required variables for compute_Pi_R (the slowest step)
        self.D_pre = []
        for D_in, cutD in enumerate(self.cutD_genomic_CDR3_segs):
            dseq = self._nt_lut[np.frombuffer(cutD.encode('ascii'), dtype=np.uint8)]
            PD_2nd_all = self.PD_2nd_nt_pos_per_aa_vec[D_in]
            PD_2nd_stack = np.empty((len(self.i2aa), 4, dseq.shape[0]))
            for aidx, aa in enumerate(self.i2aa):
                PD_2nd_stack[aidx] = PD_2nd_all[aa]

            self.D_pre.append({
                'dseq': dseq,
                'Pdel_D': self.PdelDldelDr_given_D[:, :, D_in],
                'max_del_D': np.asarray(self.max_delDl_given_DdelDr[D_in]),
                'min_del_D': np.asarray(self.min_delDl_given_DdelDr[D_in]),
                'PD_nt': self.PD_nt_pos_vec[D_in],
                'PD_2nd_stack': PD_2nd_stack,
                'zeroD': self.zeroD_given_D[D_in],
            })

        # Converting the codon lookup tables into arrays so we can access
        # in the numba functions
        num_aa = len(self.i2aa)
        self.allow1 = np.zeros((num_aa,  4), dtype=np.uint8)
        self.allow2 = np.zeros((num_aa, 16), dtype=np.uint8)
        self.allow3 = np.zeros((num_aa, 64), dtype=np.uint8)
        self.allow_lsf_stack = np.zeros((num_aa,  4, 4, 4), dtype=np.uint8)

        self.nt2i = {'A': 0, 'C': 1, 'G': 2, 'T': 3}
        for aa, aidx in self.aa2i.items():
            for sfx in self.sub_codons_right[aa]:
                if len(sfx) == 1:
                    l = self.nt2i[sfx[0]]
                    self.allow1[aidx, l] = 1
                elif len(sfx) == 2:
                    s = self.nt2i[sfx[0]]
                    l = self.nt2i[sfx[1]]
                    self.allow2[aidx, s*4 + l] = 1

        for aa, aidx in self.aa2i.items():
            for c in self.codons_dict[aa]:
                l = self.nt2i[c[0]]
                s = self.nt2i[c[1]]
                f = self.nt2i[c[2]]
                self.allow3[aidx, l*16 + s*4 + f] = 1
                self.allow_lsf_stack[aidx, l, s, f] = 1


    def __getattr__(self, name):
        return getattr(self._impl, name)

    def _warmup_numba(self):
        """Pre-compile numba functions to avoid JIT overhead in parallel workers.
        
        This method should be called after unpickling to ensure all numba functions
        are compiled before the first real computation in a worker process.
        The most effective warmup is to actually call compute_CDR3_pgen with a real
        sequence, which exercises all code paths.
        """
        try:
            # Best warmup: call the actual compute method with a simple sequence
            # This ensures all numba functions are compiled with the correct signatures
            # Use a simple sequence that's likely to work
            dummy_seq = "CA"  # Minimal valid sequence
            # Get valid V and J indices (use first available)
            if hasattr(self._impl, 'V_mask_mapping'):
                V_indices = list(self._impl.V_mask_mapping.values()) if hasattr(self._impl.V_mask_mapping, 'values') else [0]
                J_indices = list(self._impl.J_mask_mapping.values()) if hasattr(self._impl.J_mask_mapping, 'values') else [0]
            else:
                # Fallback: try to get V/J usage masks
                try:
                    # Try to compute with minimal masks
                    V_usage_mask = [1] * (len(self._impl.V_allele_names) if hasattr(self._impl, 'V_allele_names') else 1)
                    J_usage_mask = [1] * (len(self._impl.J_allele_names) if hasattr(self._impl, 'J_allele_names') else 1)
                    # Actually call compute_CDR3_pgen - this will compile everything
                    _ = self.compute_CDR3_pgen(dummy_seq, V_usage_mask, J_usage_mask)
                    return  # Success - all functions compiled
                except Exception:
                    pass
            
            # Fallback: manual warmup if the above fails
            dummy_aa_seq = self._aa_lut[np.frombuffer(dummy_seq.encode('latin-1'), np.uint8)]
            dummy_aa_seq = dummy_aa_seq[dummy_aa_seq >= 0]
            if len(dummy_aa_seq) == 0:
                return  # Can't warmup without valid sequence
            
            L3 = len(dummy_aa_seq) * 3
            
            if isinstance(self._impl, GenerationProbabilityVDJ):
                # Warm up VDJ functions
                dummy_Pi_V = np.ones((4, L3), dtype=np.float64) * 0.25
                dummy_out_L = np.zeros((4, L3))
                max_align = min(L3, 15)
                compute_Pi_L_numba(
                    dummy_out_L, dummy_aa_seq, dummy_Pi_V,
                    self.PinsVD,
                    self.first_nt_bias_insVD,
                    self.zero_nt_bias_insVD,
                    self.Svd_stack, self.Dvd_stack, self.Tvd_stack,
                    self.lDvd_stack, self.lTvd_stack,
                    max_align
                )
                
                dummy_Pi_J_given_D = [np.ones((4, L3), dtype=np.float64) * 0.25]
                dummy_out_J = np.zeros((1, 4, L3))
                compute_Pi_JinsDJ_given_D_numba(
                    dummy_out_J,
                    dummy_aa_seq,
                    np.stack(dummy_Pi_J_given_D, axis=0),
                    self.PinsDJ,
                    self.first_nt_bias_insDJ,
                    self.zero_nt_bias_insDJ,
                    self.Sdj_stack, self.Ddj_stack, self.Tdj_stack,
                    self.rDdj_stack, self.rTdj_stack,
                    max_align
                )
                
                dummy_Pi_R = np.zeros((4, L3))
                dummy_Pi_JinsDJ = np.ones((4, L3), dtype=np.float64) * 0.25
                for pre in self.D_pre[:1]:  # Just warm up first D segment
                    compute_Pi_R_one_numba(
                        dummy_Pi_R, L3, -L3,
                        pre['dseq'],
                        dummy_Pi_JinsDJ,
                        pre['Pdel_D'],
                        pre['max_del_D'],
                        pre['min_del_D'],
                        pre['PD_nt'], pre['PD_2nd_stack'], pre['zeroD'],
                        dummy_aa_seq, self.allow_lsf_stack, self.allow1, self.allow2, self.allow3
                    )
            else:
                # Warm up VJ functions
                dummy_Pi_V_given_J = [np.ones((4, L3), dtype=np.float64) * 0.25]
                dummy_out_V = np.zeros((1, 4, L3))
                max_V_align = min(L3, 15)
                compute_Pi_V_insVJ_given_J_numba(
                    dummy_out_V,
                    dummy_aa_seq,
                    np.stack(dummy_Pi_V_given_J, axis=0),
                    self.PinsVJ,
                    self.first_nt_bias_insVJ,
                    self.zero_nt_bias_insVJ,
                    self.Svj_stack, self.Dvj_stack, self.Tvj_stack,
                    self.lDvj_stack, self.lTvj_stack,
                    max_V_align
                )
        except Exception:
            # Silently ignore warmup errors - functions will compile on first use
            pass

    def __getstate__(self):
        """Custom pickling: store all instance attributes and remove monkey patch."""
        state = self.__dict__.copy()
        # Remove the monkey-patched method from _impl to avoid pickling closure issues
        # The closure in the monkey-patched method references self, which causes circular references
        # Check if it's an instance attribute (our monkey patch) vs class method
        if hasattr(self._impl, 'compute_CDR3_pgen'):
            # Check if it's stored as an instance attribute (our monkey patch)
            if 'compute_CDR3_pgen' in self._impl.__dict__:
                # Temporarily remove it - we'll restore it in __setstate__
                del self._impl.__dict__['compute_CDR3_pgen']
        return state

    def __setstate__(self, state):
        """Custom unpickling: restore all attributes and re-apply monkey patch."""
        # Restore all attributes
        self.__dict__.update(state)
        
        # Always restore the monkey patch (it's always applied in __init__)
        def _proxy(impl_self, *args, **kwargs):
            return self.compute_CDR3_pgen(*args, **kwargs)
        self._impl.compute_CDR3_pgen = MethodType(_proxy, self._impl)
        
        # Warm up numba functions to avoid JIT compilation overhead in parallel workers
        self._warmup_numba()


    def compute_CDR3_pgen(
        self,
        CDR3_seq: str,
        V_usage_mask: List[int],
        J_usage_mask: List[int]
    ) -> float:

        if isinstance(self._impl, GenerationProbabilityVDJ):
            Pi_V, max_V_align = self._impl.compute_Pi_V(CDR3_seq, V_usage_mask)
            Pi_L = self.compute_Pi_L(CDR3_seq, Pi_V, max_V_align)

            Pi_J_given_D, max_J_align = self._impl.compute_Pi_J_given_D(CDR3_seq, J_usage_mask)
            Pi_JinsDJ_given_D = self.compute_Pi_JinsDJ_given_D(CDR3_seq, Pi_J_given_D, max_J_align)
            Pi_R = self.compute_Pi_R(CDR3_seq, Pi_JinsDJ_given_D)
   
            pgen = np.sum(Pi_L[:, :-1] * Pi_R[:, 1:])
        else:
            Pi_J, r_J_usage_mask = self._impl.compute_Pi_J(CDR3_seq, J_usage_mask)
            if len(Pi_J) == 0:
                return 0.0
            
            Pi_V_given_J, max_V_align = self._impl.compute_Pi_V_given_J(CDR3_seq, V_usage_mask, r_J_usage_mask)
            if len(Pi_V_given_J) == 0:
                return 0.0
            
            Pi_V_insVJ_given_J = self.compute_Pi_V_insVJ_given_J(CDR3_seq, Pi_V_given_J, max_V_align)    
            Pi_J = np.stack(Pi_J)
            pgen = np.dot(Pi_V_insVJ_given_J[:, :, :-1].ravel(), Pi_J[:, :, 1:].ravel())

        return pgen


    def compute_Pi_V_insVJ_given_J(
        self,
        CDR3_seq: str,
        Pi_V_given_J: List[np.ndarray],
        max_V_align: int
    ) -> np.ndarray:
        L3 = 3 * len(CDR3_seq)
        numD = len(Pi_V_given_J)

        out = np.zeros((numD, 4, L3))
        Pi_V_given_J = np.stack(Pi_V_given_J, axis=0)
        aa_idx_seq = self._aa_lut[np.frombuffer(CDR3_seq.encode('latin-1'), np.uint8)]
        
        compute_Pi_V_insVJ_given_J_numba(
            out,
            aa_idx_seq,
            Pi_V_given_J,
            self.PinsVJ,
            self.first_nt_bias_insVJ,
            self.zero_nt_bias_insVJ,
            self.Svj_stack, self.Dvj_stack, self.Tvj_stack,
            self.lDvj_stack, self.lTvj_stack,
            max_V_align
        )
        return out


    def compute_Pi_JinsDJ_given_D(
        self,
        CDR3_seq: str,
        Pi_J_given_D: List[np.ndarray],
        max_J_align: int
    ) -> np.ndarray:
        L3 = 3 * len(CDR3_seq)
        numD = len(Pi_J_given_D)

        out = np.zeros((numD, 4, L3))
        Pi_J_given_D = np.stack(Pi_J_given_D, axis=0)
        aa_idx_seq = self._aa_lut[np.frombuffer(CDR3_seq.encode('latin-1'), np.uint8)]

        compute_Pi_JinsDJ_given_D_numba(
            out,
            aa_idx_seq,
            Pi_J_given_D,
            self.PinsDJ,
            self.first_nt_bias_insDJ,
            self.zero_nt_bias_insDJ,
            self.Sdj_stack, self.Ddj_stack, self.Tdj_stack,
            self.rDdj_stack, self.rTdj_stack,
            max_J_align
        )
        return out
 

    def compute_Pi_R(self, CDR3_seq: str, Pi_JinsDJ_given_D: np.ndarray) -> np.ndarray:
        L3 = len(CDR3_seq) * 3
        Pi_R = np.zeros((4, L3))
        aa_idx_seq = self._aa_lut[np.frombuffer(CDR3_seq.encode('latin-1'), np.uint8)]

        allow_lsf_stack = self.allow_lsf_stack
        allow1 = self.allow1
        allow2 = self.allow2
        allow3 = self.allow3

        for D_in, pre in enumerate(self.D_pre):
            compute_Pi_R_one_numba(
                Pi_R, L3, -L3,
                pre['dseq'],
                Pi_JinsDJ_given_D[D_in],
                pre['Pdel_D'],
                pre['max_del_D'],
                pre['min_del_D'],
                pre['PD_nt'], pre['PD_2nd_stack'], pre['zeroD'],
                aa_idx_seq, allow_lsf_stack, allow1, allow2, allow3
            )

        return Pi_R
    

    def compute_Pi_L(self, CDR3_seq: str, Pi_V: np.ndarray, max_V_align: int) -> np.ndarray:
        L3 = len(CDR3_seq) * 3
        Pi_L = np.zeros((4, L3))
        aa_idx_seq = self._aa_lut[np.frombuffer(CDR3_seq.encode('latin-1'), np.uint8)]

        compute_Pi_L_numba(
            Pi_L, aa_idx_seq, Pi_V,
            self.PinsVD,
            self.first_nt_bias_insVD,
            self.zero_nt_bias_insVD,
            self.Svd_stack, self.Dvd_stack, self.Tvd_stack,
            self.lDvd_stack, self.lTvd_stack,
            max_V_align
        )
        return Pi_L



#!/usr/bin/env python3
import sys
import math
import struct
from pathlib import Path
from functools import reduce

import click
import polars as pl #v0.18.0
import numpy as np


def ulp_distance_1(a: float, b: float) -> int:
    a, b = sorted((a, b))
    cur = a
    ulps = 0
    while cur != b:
        nex = np.nextafter(cur, b)
        ulps += 1
        cur = nex
    
    return ulps


def ulp_distance_2(a: float, b: float) -> int:
    if math.isnan(a) or math.isnan(b):
        return math.inf

    ai = struct.unpack('>q', struct.pack('>d', a))[0]
    bi = struct.unpack('>q', struct.pack('>d', b))[0]

    if ai < 0:
        ai = (0x8000000000000000 - ai) & 0xFFFFFFFFFFFFFFFF
    if bi < 0:
        bi = (0x8000000000000000 - bi) & 0xFFFFFFFFFFFFFFFF

    return abs(ai - bi)


@click.command()
@click.argument('files', type=click.Path(exists=True), nargs=-1, required=True)
def main(files):
    """ Compare old pgen results vs new by calculating the ULP distance.

    Input is a 2-col csv containing the old Pgen and new Pgen scores.
    (can be built from from `xsv cat columns old.csv new.csv`)
    """

    hists = []
    for file in files:
        label = Path(file).stem

        df = pl.read_csv(
            file,
            has_header=False,
            new_columns=['old', 'new']
        )

        # We'll use 2 implementations of ULP distance in case one is faulty
        df = (
            df.with_columns([
                pl.struct(['old', 'new'])
                    .apply(lambda x: ulp_distance_1(x['old'], x['new']))
                    .alias('ulp_distance'),
                pl.struct(['old', 'new'])
                    .apply(lambda x: ulp_distance_2(x['old'], x['new']))
                    .alias('ulp_distance_2'),
            ])
        )

        if not df.filter(pl.col('ulp_distance') != pl.col('ulp_distance_2')).is_empty():
            raise RuntimeError('1 of the ULP distance functions is off.')

        df_hist = (
            df
            .groupby('ulp_distance')
            .count()
            .sort('ulp_distance')
            .with_columns(
                pl.col('count').cumsum().alias('cumulative')
            )
            .with_columns(
                (pl.col('cumulative') / pl.col('cumulative').max()) * 100.0
            )
            .select(['ulp_distance', 'count'])
            .rename({'count': label})
        )
        hists.append(df_hist)

    #print (df_agg)
    
    df_hist = reduce(lambda l, r: l.join(r, on='ulp_distance', how='outer'), hists)
    df_hist.write_csv(sys.stdout.buffer)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Module for utility functions.

    Copyright (C) 2018 Zachary Sethna

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    
    
For documentation on how to define custom amino acid alphabets, look at the
docstring of construct_codons_dict.

@author: zacharysethna
"""
from __future__ import print_function, division
import numpy as np

    
#%% Translation functions
def nt2aa(ntseq):
    """Translate a nucleotide sequence into an amino acid sequence.

    Parameters
    ----------
    ntseq : str
        Nucleotide sequence composed of A, C, G, or T (uppercase or lowercase)

    Returns
    -------
    aaseq : str
        Amino acid sequence
    
    Example
    --------
    >>> nt2aa('TGTGCCTGGAGTGTAGCTCCGGACAGGGGTGGCTACACCTTC')
    'CAWSVAPDRGGYTF'
        
    """
    nt2num = {'A': 0, 'C': 1, 'G': 2, 'T': 3, 'a': 0, 'c': 1, 'g': 2, 't': 3}
    aa_dict ='KQE*TPASRRG*ILVLNHDYTPASSRGCILVFKQE*TPASRRGWMLVLNHDYTPASSRGCILVF'
    
    return ''.join([aa_dict[nt2num[ntseq[i]] + 4*nt2num[ntseq[i+1]] + 16*nt2num[ntseq[i+2]]] for i in range(0, len(ntseq), 3) if i+2 < len(ntseq)])

def nt2codon_rep(ntseq):
    """Represent nucleotide sequence by sequence of codon symbols.

    'Translates' the nucleotide sequence into a symbolic representation of
    'amino acids' where each codon gets its own unique character symbol. These 
    characters should be reserved only for representing the 64 individual 
    codons --- note that this means it is important that this function matches 
    the corresponding function in the preprocess script and that any custom 
    alphabet does not use these symbols. Defining symbols for each individual
    codon allows for Pgen computation of inframe nucleotide sequences. 

    Parameters
    ----------

    ntseq : str
        A Nucleotide sequence (normally a CDR3 nucleotide sequence) to be 
        'translated' into the codon - symbol representation. Can be either
        uppercase or lowercase, but only composed of A, C, G, or T.

    Returns
    -------
    codon_rep : str
        The codon - symbolic representation of ntseq. Note that if 
        len(ntseq) == 3L --> len(codon_rep) == L
    
    Example
    --------
    >>> nt2codon_rep('TGTGCCTGGAGTGTAGCTCCGGACAGGGGTGGCTACACCTTC')
    '\xbb\x96\xab\xb8\x8e\xb6\xa5\x92\xa8\xba\x9a\x93\x94\x9f'
        
    """
    
    #
    nt2num = {'A': 0, 'C': 1, 'G': 2, 'T': 3, 'a': 0, 'c': 1, 'g': 2, 't': 3}
    #Use single characters not in use to represent each individual codon --- this function is called in constructing the codon dictionary
    codon_rep ='\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf\xb0\xb1\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf'
    
    return ''.join([codon_rep[nt2num[ntseq[i]] + 4*nt2num[ntseq[i+1]] + 16*nt2num[ntseq[i+2]]] for i in range(0, len(ntseq), 3) if i+2 < len(ntseq)])

#%% Functions that delete/insert reverse palindromes nucleotide sequences
def cutR_seq(seq, cutR, max_palindrome):
    """Cut genomic sequence from the right.

    Parameters
    ----------
    seq : str
        Nucleotide sequence to be cut from the right
    cutR : int
        cutR - max_palindrome = how many nucleotides to cut from the right.
        Negative cutR implies complementary palindromic insertions.
    max_palindrome : int
        Length of the maximum palindromic insertion.

    Returns
    -------
    seq : str
        Nucleotide sequence after being cut from the right
    
    Examples
    --------
    >>> cutR_seq('TGCGCCAGCAGTGAGTC', 0, 4)
    'TGCGCCAGCAGTGAGTCGACT'
    >>> cutR_seq('TGCGCCAGCAGTGAGTC', 8, 4)
    'TGCGCCAGCAGTG'
    
    """
    complement_dict = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A'} #can include lower case if wanted
    if cutR < max_palindrome:
        seq = seq + ''.join([complement_dict[nt] for nt in seq[cutR - max_palindrome:]][::-1]) #reverse complement palindrome insertions
    else:
        seq = seq[:len(seq) - cutR + max_palindrome] #deletions
    
    return seq
    
def cutL_seq(seq, cutL, max_palindrome):
    """Cut genomic sequence from the left.

    Parameters
    ----------
    seq : str
        Nucleotide sequence to be cut from the right
    cutL : int
        cutL - max_palindrome = how many nucleotides to cut from the left.
        Negative cutL implies complementary palindromic insertions.
    max_palindrome : int
        Length of the maximum palindromic insertion.

    Returns
    -------
    seq : str
        Nucleotide sequence after being cut from the left
    
    Examples
    --------
    >>> cutL_seq('TGAACACTGAAGCTTTCTTT', 8, 4)
    'CACTGAAGCTTTCTTT'
    >>> cutL_seq('TGAACACTGAAGCTTTCTTT', 0, 4)
    'TTCATGAACACTGAAGCTTTCTTT'
    
    """
    
    complement_dict = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A'} #can include lower case if wanted
    if cutL < max_palindrome:
        seq = ''.join([complement_dict[nt] for nt in seq[:max_palindrome - cutL]][::-1]) + seq #reverse complement palindrome insertions
    else:
        seq = seq[cutL-max_palindrome:] #deletions
    
    return seq

#%% Define codon dicts that define the 'amino acid' alphabet
def construct_codons_dict(alphabet_file = None):
    """Generate the sub_codons_right dictionary of codon suffixes.

    syntax of custom alphabet_files:
        
    char: list,of,amino,acids,or,codons,separated,by,commas
    
    
    Parameters
    ----------
    alphabet_file : str
        File name for a custom alphabet definition. If no file is provided, the 
        default alphabet is used, i.e. standard amino acids, undetermined amino 
        acids (B, J, X, and Z), and single codon symbols.

    Returns
    -------
    codons_dict : dict
        Dictionary, keyed by the allowed 'amino acid' symbols with the values 
        being lists of codons corresponding to the symbol.
    
    """
    
    #Some symbols can't be used in the CDR3 sequences in order to allow for
    #regular expression parsing and general manipulation.
    protected_symbols = [' ', '\t', '\n', '\x0b', '\x0c', '\r', ':', ',', ';', '[', ']', '{', '}', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
    
    
    #construct list of all 64 codons
    codons = [i + j + k for i in 'ACGT' for j in 'ACGT' for k in 'ACGT']
    
    codons_dict = {}
    #add standard amino acids symbols to the dict (i.e. 'ACDEFGHIKLMNPQRSTVWY*').
    #these symbols CANNOT be overwritten by custom alphabet files
    for codon in codons:
        codons_dict[nt2aa(codon)] = codons_dict.get(nt2aa(codon), []) + [codon]
                 
    #add single codon symbols to allow for inframe ntseq pgen computation
    #'\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf\xb0\xb1\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf'
    #these symbols CANNOT be overwritten by custom alphabet files
    for codon in codons:
        codons_dict[nt2codon_rep(codon)] = [codon]  
    
    #Check to see if custom alphabet file is supplied, else use default alphabet
    
    #Include standard ambigious amino acids.
    #these symbols CAN be overwritten by custom alphabet files
    expanded_alphabet = {}
    expanded_alphabet['B'] = ['D','N']
    expanded_alphabet['J'] = ['I', 'L']
    expanded_alphabet['X'] = ['A','C','D','E','F','G','H','I','K','L','M','N','P','Q','R','S','T','V','W','Y']
    expanded_alphabet['Z'] = ['E', 'Q']
    if alphabet_file is not None: #Use custom alphabet file definitions
        alphabet_f = open(alphabet_file, 'r')
        for line in alphabet_f:
            #assumed syntax is of a line is:
            #s: a1, a2, a3, a4, a5, ..., aN
            #where s is a single character symbol that isn't reserved, and all 
            #of the a's are either amino acid symbols or codons. Whitespaces
            #will be stripped as will brackets if the a's are presented as a
            #list.
            c_symbol = line.split(':', 1)[0].strip(''.join(protected_symbols))
            #Note there shouldn't be any additional colons -- this is a protected symbol.
            c_aa_codon_list_str = line.split(':', 1)[1]
            expanded_alphabet[c_symbol] = [x.strip(''.join(protected_symbols)) for x in c_aa_codon_list_str.split(',')]
        
        alphabet_f.close()
    
    
    for symbol in expanded_alphabet.keys():
        #Double check that the symbol isn't already used (important particularly for the single codon representation)
        if symbol in codons_dict.keys():
            print(symbol + " is already used as an 'amino acid' symbol for codons: ")
            print(codons_dict[symbol])
            continue
        elif not len(symbol) == 1: #Check that the custom symbol is a single character
            print("Can't use " + symbol + " as a custom 'amino acid' definitions as such symbols must be single characters.")
            continue
        elif symbol in protected_symbols: #This elif shouldn't trigger due to the stripping of protected symbols.
            print(symbol + " is a protected character")
        current_codon_collection = set()
        for x in expanded_alphabet[symbol]:
            if x in codons_dict.keys(): #Check if reference to an amino acid or other amino acid symbol
                current_codon_collection = current_codon_collection.union(codons_dict[x]) #If so, add those codons to the new collection
            elif x.upper() in codons: #Check if specifying a single codon
                current_codon_collection.add(x.upper()) #If so, add the codon to the new collection
            elif len(x) == 0: #fully stripped away
                continue
            else: #If not, don't recognize the addition and continue.
                print('Unfamiliar amino acid symbol or codon: ' + x)
                continue
        codons_dict[symbol] = list(current_codon_collection)
    
                
    return codons_dict

def generate_sub_codons_left(codons_dict):
    """Generate the sub_codons_left dictionary of codon prefixes.

    Parameters
    ----------
    codons_dict : dict
        Dictionary, keyed by the allowed 'amino acid' symbols with the values 
        being lists of codons corresponding to the symbol.

    Returns
    -------
    sub_codons_left : dict
        Dictionary of the 1 and 2 nucleotide prefixes (read from 5') for 
        each codon in an 'amino acid' grouping
        
    """
    sub_codons_left = {}
    for aa in codons_dict.keys():
        sub_codons_left[aa] = list(set([x[0] for x in codons_dict[aa]] + [x[:2] for x in codons_dict[aa]]))
        
    return sub_codons_left
    
def generate_sub_codons_right(codons_dict):
    """Generate the sub_codons_right dictionary of codon suffixes.

    Parameters
    ----------
    codons_dict : dict
        Dictionary, keyed by the allowed 'amino acid' symbols with the values 
        being lists of codons corresponding to the symbol.

    Returns
    -------
    sub_codons_right : dict
        Dictionary of the 1 and 2 nucleotide suffixes (read from 5') for 
        each codon in an 'amino acid' grouping.
        
    """
    sub_codons_right = {}
    for aa in codons_dict.keys():
        sub_codons_right[aa] = list(set([x[-1] for x in codons_dict[aa]] + [x[-2:] for x in codons_dict[aa]]))
        
    return sub_codons_right        

def determine_seq_type(seq, aa_alphabet):
    """Determine the type of a sequence.
    
    Parameters
    ----------

    seq : str
        Sequence to be typed.
    aa_alphabet : str
        String of all characters recoginized as 'amino acids'. (i.e. the keys
        of codons_dict: aa_alphabet = ''.join(codons_dict.keys())  )

    Returns
    -------
    seq_type : str
        The type of sequence (ntseq, aaseq, regex, None) seq is.
    
    Example
    --------
    >>> determine_seq_type('TGTGCCAGCAGTTCCGAAGGGGCGGGAGGGCCCTCCCTGAGAGGTCATGAGCAGTTCTTC', aa_alphabet)
    'ntseq'
    >>> determine_seq_type('CSARDX[TV]GNX{0,}', aa_alphabet)
    'regex
    
    """
    
    if all([x in 'ACGTacgt' for x in seq]):
        return 'ntseq'
    elif all([x in aa_alphabet for x in seq]):
        return 'aaseq'
    elif all([x in aa_alphabet + '[]{}0123456789,' for x in seq]):
        return 'regex'
    
#%%
#If using the steady-state distribution for first nucleotide probabilities we include a function to compute it
def calc_steady_state_dist(R):
    """Calculate the steady state dist of a 4 state markov transition matrix.

    Parameters
    ----------
    R : ndarray
        Markov transition matrix

    Returns
    -------
    p_ss : ndarray
        Steady state probability distribution
        
    """
    #Calc steady state distribution for a dinucleotide bias matrix
    
    w, v = np.linalg.eig(R)
    
    for i in range(4):
        if np.abs(w[i] - 1) < 1e-8:
            return np.real(v[:, i] / np.sum(v[:, i]))
    return -1

#%% Gene recognition
def gene_to_num_str(gene_name, gene_type):
    """Strips excess gene name info to number string.
    
    Parameters
    ----------
    gene_name : str
        Gene or allele name
    gene_type : char
        Genomic cassette type. (i.e. V, D, or J)

    Returns
    -------
    num_str : str
        Reduced gene or allele name with leading zeros and excess
        characters removed.
        
    """

    num_str = gene_type.lower().join([g.lstrip('0') for g in gene_name.lower().split(gene_type.lower())[1:]])
    num_str = '-'.join([g.lstrip('0') for g in num_str.split('-')])
    num_str = '*'.join([g.lstrip('0') for g in num_str.split('*')])
    return gene_type.lower() + num_str.replace('/', '') # get rid of '/' too

#%% Entropy Functions
def calc_S(P, base):
    """Calculate the entropy of a probability distribution in base units.
    
    Parameters
    ----------
    P : ndarray
        Probability distribution
    base : float
        Base used to set entropy units. Suggestions:
        base = 2 (entropy in bits)
        base = e (entropy in nats)
        base = 10 (entropy in dits)

    Returns
    -------
    S : float
        Entropy of P
        
    """
    return np.sum([-p*np.log(p) for p in P.flatten() if p > 0])/np.log(base)

def calc_S_single_gene(PG, PdelG_given_G, base = 2):
    """Calculate the entropy of an independent gene usage and deletion dists.
    
    Parameters
    ----------
    PG : ndarray
        Probability distribution of the usage of G-type genes (e.g PV)
    PdelG_given_G : ndarray
        Deletion profiles of each G gene conditioned on the gene 
        (e.g. PdelV_given_V or PdelDldelDr_given_D)
    base : float
        Base used to set entropy units. Suggestions:
        base = 2 (entropy in bits, default)
        base = e (entropy in nats)
        base = 10 (entropy in dits)

    Returns
    -------
    SG : float
        Entropy of PG (entropy of gene choice)
    SdelG : float
        Conditional entropy of PdelG_given_G (entropy of deletions)
        
    """
    
    SG = calc_S(PG, base)
    SdelG = 0
    for g, pg in enumerate(PG):
        if pg == 0:
            continue
        SdelG += pg * calc_S(PdelG_given_G[..., g], base)
        
    return SG, SdelG

def calc_S_joint_genes(PG1G2, PdelG1_given_G1, PdelG2_given_G2, base = 2):
    """Calculate the entropy of a joint gene usage and deletion dists.
    
    Parameters
    ----------
    PG1G2 : ndarray
        Joint probability distribution of the usage of G1-type genes and 
        G2-type genes (e.g. PDJ or PVJ)
    PdelG1_given_G1 : ndarray
        Deletion profiles of each G1 gene conditioned on the gene 
        (e.g. PdelV_given_V or PdelDldelDr_given_D)
    PdelG2_given_G2 : ndarray
        Deletion profiles of each G2 gene conditioned on the gene 
        (e.g. PdelV_given_V or PdelDldelDr_given_D)
    base : float
        Base used to set entropy units. Suggestions:
        base = 2 (entropy in bits, default)
        base = e (entropy in nats)
        base = 10 (entropy in dits)

    Returns
    -------
    SG1G2 : float
        Entropy of PG1G2 (entropy of gene choice)
    SdelG1 : float
        Conditional entropy of PdelG1_given_G1 (entropy of G1 deletions)
    SdelG2 : float
        Conditional entropy of PdelG2_given_G2 (entropy of G2 deletions)
        
    """

    PG1 = np.sum(PG1G2, axis = 1)
    PG2 = np.sum(PG1G2, axis = 0)
    
    SG1G2 = calc_S(PG1G2, base)
    SdelG1 = 0
    for g1, pg1 in enumerate(PG1):
        if pg1 == 0:
            continue
        SdelG1 += pg1 * calc_S(PdelG1_given_G1[..., g1], base)

    SdelG2 = 0
    for g2, pg2 in enumerate(PG2):
        if pg2 == 0:
            continue
        SdelG2 += pg2 * calc_S(PdelG2_given_G2[..., g2], base)
        
    return SG1G2, SdelG1, SdelG2

def calc_Sins(Pins, R, p_0 = None, base = 2):
    """Calculate the entropy of an insertion junction.

    Entropy of nucleotide choice in the Markov model is given by:
    H(p_0)(1-Pins[0]) + \sum_{l=2}Pins[l] \sum_{k=2}^l \sum_m R^{k-1}p_0(m) H(R[:, m])
    
    
    Parameters
    ----------
    Pins : ndarray
        Length distribution of the junction (e.g. PinsVD, PinsDJ, or PinsVJ)
    R : ndarray
        Markov transition matrix
    base : float
        Base used to set entropy units. Suggestions:
        base = 2 (entropy in bits, default)
        base = e (entropy in nats)
        base = 10 (entropy in dits)

    Returns
    -------
    Sins : float
        Entropy of Pins (entropy of length distribution)
    Smarkov : float
        Entropy of nucleotide choice in the Markov model.
        
    """
    Sins = calc_S(Pins, base)
    if p_0 is None: #Use steady state if no p_0 given
        p_0 = calc_steady_state_dist(R)
    cr_Pins = Pins[::-1].cumsum()[::-1]
    
    Smarkov = calc_S(p_0, base)*(1-Pins[0])
    p_k = p_0
    for k in range(2, len(Pins)):
        p_k = np.dot(R, p_k)
        Smarkov += cr_Pins[k] * np.sum([p_k[m]*calc_S(R[:, m], base) for m in range(4)])
    
    return Sins, Smarkov


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Process classes GenerativeModelV(D)J GenomicModelV(D)J before computing Pgen.

    Copyright (C) 2018 Zachary Sethna

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    

As many of the arrays that will be used to compute the Pgen of a specific 
sequence depend only on the generative model and the germline sequences (as 
opposed to the CDR3 sequence itself) much of the computation and organization
can be done in advance and does not need to be repeated for each individual 
sequence. This module prepares and formats all of the sequence-independent 
components which are stored as attributes of classes PreprocessedParametersVDJ 
and PreprocessedParametersVJ which will be inherited by the 
GenerationProbabilityV(D)J classes.

There is also an option of defining a custom 'amino acid' alphabet. Please 
refer to the example alphabet file and the README for more info on how to 
define customized alphabets for faster Pgen computation of CDR3 motifs.

This module and the classes PreprocessedParametersV(D)J are unlikely to be
used directly, but are used in the instantiation of the classes
GenerationProbabilityV(D)J.

@author: zacharysethna

"""
from __future__ import division
import numpy as np
from olga.utils import construct_codons_dict, generate_sub_codons_left, generate_sub_codons_right, calc_steady_state_dist, gene_to_num_str

class PreprocessedParameters(object):
    """Class used to preprocess the parameters that both VDJ and VJ models have.

    The class is the parent of classes PreprocessedParametersVDJ and 
    PreprocessedParametersVJ.

    Attributes
    ----------
    codons_dict : dict
        Dictionary, keyed by the allowed 'amino acid' symbols with the values 
        being lists of codons corresponding to the symbol.        
    sub_codons_left : dict
        Dictionary of the 1 and 2 nucleotide prefixes (read from 5') for 
        each codon in an 'amino acid' grouping        
    sub_codons_right : dict
        Dictionary of the 1 and 2 nucleotide suffixes (read from 5') for 
        each codon in an 'amino acid' grouping
        
    V_allele_names : list of strings
        List of V allele names in genomic_data        
    d_V_usage_mask : list of int
        Default V usage mask of indices of all productive V genes/alleles.        
    V_mask_mapping : dict
        Dictionary mapping allowed keywords (strings corresponding to V gene 
        and V allele names) to the indices of the V alleles they refer to.        
    cutV_genomic_CDR3_segs : list of strings
        List of the V germline nucleotide sequences, trimmed to begin at the
        CDR3 region (includes the conserved C residue) with the maximum number
        of reverse complementary palindromic insertions appended.
        
    J_allele_names : list of strings
        List of J allele names in genomic_data        
    d_J_usage_mask : list of int
        Default J usage mask of indices of all productive J genes/alleles.        
    J_mask_mapping : dict
        Dictionary mapping allowed keywords (strings corresponding to V gene 
        and V allele names) to the indices of the V alleles they refer to.        
    cutJ_genomic_CDR3_segs : list of strings
        List of the J germline nucleotide sequences, trimmed to end at the
        CDR3 region (includes the conserved F or W residue) with the maximum 
        number of reverse complementary palindromic insertions appended.
        
    """
    def __init__(self, generative_model, genomic_data, alphabet_file = None):
        """Initialize PreprocessedParameters
        
        This initialization is used to compute the preprocess parameters that 
        are in common with both the VDJ and VJ model types.
        
        Parameters
        ----------
        generative_model : GenerativeModelVDJ or GenerativeModelVJ
            V(D)J generative model classes containing the model parameters.            
        genomic_data : GenomicDataVDJ GenomicDataVJ
            V(D)J genomic data classes containing the V(,D,) and J germline 
            sequences and info.            
        alphabet_file : str, optional
            File name (full pathing from current directory) for a custom alphabet
            definition. If no file is provided, the default alphabet is used, i.e. 
            standard amino acids, undetermined amino acids (B, J, X, and Z), and
            single codon symbols.
        
        """
        
        #determine if generative_model and genomic_data are VDJ or VJ
        if generative_model.__class__.__name__.endswith('VDJ') and genomic_data.__class__.__name__.endswith('VDJ'):
            recomb_type = 'VDJ'
        elif generative_model.__class__.__name__.endswith('VJ') and genomic_data.__class__.__name__.endswith('VJ'):
            recomb_type = 'VJ'
        else:
            raise ValueError #recomb types must match
        
        
        self.codons_dict = construct_codons_dict(alphabet_file)
        self.sub_codons_left = generate_sub_codons_left(self.codons_dict)
        self.sub_codons_right = generate_sub_codons_right(self.codons_dict)
        
        self.V_allele_names = None
        self.V_mask_mapping = None
        self.J_allele_names = None
        self.J_mask_mapping = None
        self.make_V_and_J_mask_mapping(genomic_data.genV, genomic_data.genJ)
        
        self.cutV_genomic_CDR3_segs = genomic_data.cutV_genomic_CDR3_segs
        
        #construct the default V_usage_mask which represents all functional V genes with non-zero probability
        d_V_usage_mask = []
        for V_in, cutVseg in enumerate(genomic_data.cutV_genomic_CDR3_segs):
            if recomb_type == 'VDJ':
                if len(cutVseg) > 0 and generative_model.PV[V_in] > 0:
                    d_V_usage_mask.append(V_in)
            elif recomb_type == 'VJ':
                if len(cutVseg) > 0 and np.sum(generative_model.PVJ[V_in, :]) > 0:
                    d_V_usage_mask.append(V_in)
        self.d_V_usage_mask = d_V_usage_mask
        
        
        
        self.cutJ_genomic_CDR3_segs = genomic_data.cutJ_genomic_CDR3_segs
        
        #construct the default J_usage_mask which represents all functional J genes with non-zero probability
        d_J_usage_mask = []
        for J_in, cutJseg in enumerate(genomic_data.cutJ_genomic_CDR3_segs):
            if recomb_type == 'VDJ':
                if len(cutJseg) > 0 and np.sum(generative_model.PDJ[:, J_in]) > 0:
                    d_J_usage_mask.append(J_in)
            elif recomb_type == 'VJ':
                if len(cutJseg) > 0 and np.sum(generative_model.PVJ[:, J_in]) > 0:
                    d_J_usage_mask.append(J_in)
        self.d_J_usage_mask = d_J_usage_mask
        
        
    def make_V_and_J_mask_mapping(self, genV, genJ):
        """Constructs the V and J mask mapping dictionaries.
        
        Parameters
        ----------
        genV : list
            List of genomic V information.            
        genJ : list
            List of genomic J information.
        
        """
        #construct mapping between allele/gene names and index for custom V_usage_masks
        V_allele_names = [V[0] for V in genV]
        V_reduced_allele_names = [gene_to_num_str(v, 'v') for v in V_allele_names]
        V_mask_mapping = {}
        for v in [x.split('*')[0] for x in V_reduced_allele_names]:
            V_mask_mapping[v] = []
            if '-' not in v:
                V_mask_mapping[v+'-1'] = []
            else:
                V_mask_mapping[v.split('-')[0]] = []

        for i, v in enumerate(V_reduced_allele_names):
            V_mask_mapping[v] = [i]
            V_mask_mapping[v.split('*')[0]].append(i)
            if '-' not in v:
                V_mask_mapping[v.replace('*', '-1*')] = [i]
                V_mask_mapping[v.split('*')[0] + '-1'].append(i)
            else:
                V_mask_mapping[v.split('*')[0].split('-')[0]].append(i)

        #construct mapping between allele/gene names and index for custom J_usage_masks
        J_allele_names = [J[0] for J in genJ]
        J_reduced_allele_names = [gene_to_num_str(j, 'j') for j in J_allele_names]
        J_mask_mapping = {}
        for j in [x.split('*')[0] for x in J_reduced_allele_names]:
            J_mask_mapping[j] = []
            if '-' not in j:
                J_mask_mapping[j+'-1'] = []
            else:
                J_mask_mapping[j.split('-')[0]] = []
        for i, j in enumerate(J_reduced_allele_names):
            J_mask_mapping[j] = [i]
            J_mask_mapping[j.split('*')[0]].append(i)
            if '-' not in j:
                J_mask_mapping[j.replace('*', '-1*')] = [i]
                J_mask_mapping[j.split('*')[0] + '-1'].append(i)
            else:
                J_mask_mapping[j.split('*')[0].split('-')[0]].append(i)
            
        self.V_allele_names = V_allele_names
        self.V_mask_mapping = V_mask_mapping
        self.J_allele_names = J_allele_names
        self.J_mask_mapping = J_mask_mapping

#%% PreprocessedParametersVJ Class def
class PreprocessedParametersVDJ(PreprocessedParameters):
    """Class used to preprocess the parameters of VDJ recombination models.

    Attributes
    ----------
    codons_dict : dict
        Dictionary, keyed by the allowed 'amino acid' symbols with the values 
        being lists of codons corresponding to the symbol.        
    sub_codons_left : dict
        Dictionary of the 1 and 2 nucleotide prefixes (read from 5') for 
        each codon in an 'amino acid' grouping        
    sub_codons_right : dict
        Dictionary of the 1 and 2 nucleotide suffixes (read from 5') for 
        each codon in an 'amino acid' grouping
        
    V_allele_names : list of strings
        List of V allele names in genomic_data        
    d_V_usage_mask : list of int
        Default V usage mask of indices of all productive V genes/alleles.        
    V_mask_mapping : dict
        Dictionary mapping allowed keywords (strings corresponding to V gene 
        and V allele names) to the indices of the V alleles they refer to.        
    cutV_genomic_CDR3_segs : list of strings
        List of the V germline nucleotide sequences, trimmed to begin at the
        CDR3 region (includes the conserved C residue) with the maximum number
        of reverse complementary palindromic insertions appended.        
    PVdelV_nt_pos_vec : list of ndarrays
        For each V allele, format P(V)*P(delV|V) into the correct form for a Pi 
        array or V_{x_1}. This is only done for the first and last position in 
        each codon.         
    PVdelV_2nd_nt_pos_per_aa_vec : list of dicts
        For each V allele, and each 'amino acid', format P(V)*P(delV|V) for 
        positions in the middle of a codon into the correct form for a Pi array 
        or V_{x_1} given the 'amino acid'.
        
    D_allele_names : list of strings
        List of D allele names in genomic_data        
    cutD_genomic_CDR3_segs : list of strings
        List of the D germline nucleotide sequences, with the maximum number
        of reverse complementary palindromic insertions appended to both ends.        
    PD_nt_pos_vec : list of ndarrays
        For each D allele, format P(delDl, delDr|D) into the correct form for a 
        Pi array as if each position were the first in a codon.    
    PD_2nd_nt_pos_per_aa_vec : list of dicts
        For each D allele, and each 'amino acid', format P(delDl, delDr|D) for 
        positions in the middle of a codon into the correct form for a Pi array 
        as if each position were the middle of a codon corresponding to the 
        'amino acid'.        
    min_delDl_given_DdelDr : list of lists
        minimum delDl for each delDr, D combination.        
    max_delDl_given_DdelDr : list of lists
        maximum delDl for each delDr, D combination.
    zeroD_given_D : list of floats
        The probability that a given D allele is fully deleted away.        
    PdelDldelDr_given_D : ndarray
        Joint probability distribution of the D deletions given the D allele,
        i.e. P(delDl, delDr |D)
        
    J_allele_names : list of strings
        List of J allele names in genomic_data        
    d_J_usage_mask : list of int
        Default J usage mask of indices of all productive J genes/alleles.        
    J_mask_mapping : dict
        Dictionary mapping allowed keywords (strings corresponding to V gene 
        and V allele names) to the indices of the V alleles they refer to.        
    cutJ_genomic_CDR3_segs : list of strings
        List of the J germline nucleotide sequences, trimmed to end at the
        CDR3 region (includes the conserved F or W residue) with the maximum 
        number of reverse complementary palindromic insertions appended.        
    PJdelJ_nt_pos_vec : list
        For each J allele, format P(J)*P(delJ|J) into the correct form for a Pi 
        array or J(D)^{x_4}. This is only done for the first and last position 
        in each codon.        
    PJdelJ_2nd_nt_pos_per_aa_vec : list
        For each J allele, and each 'amino acid', format P(J)*P(delJ|J) for 
        positions in the middle of a codon into the correct form for a Pi array 
        or J(D)^{x_4} given the 'amino acid'.
    PD_given_J : ndarray
        Probability distribution of D conditioned on J, i.e. P(D|J). 
         
    PinsVD : ndarray
        Probability distribution of the VD (N1) insertion sequence length        
    PinsDJ : ndarray
        Probability distribution of the DJ (N2) insertion sequence length        
    Rvd : ndarray
        Markov transition matrix for the VD insertion junction.
    Rdj : ndarray
        Markov transition matrix for the DJ insertion junction.        
    first_nt_bias_insVD : ndarray
        (4,) array of the probability distribution of the indentity of the 
        first nucleotide insertion for the VD junction.        
    zero_nt_bias_insVD : ndarray
        (4,) array of the probability distribution of the indentity of the 
        the nucleotide BEFORE the VD insertion.
        zero_nt_bias_insVD = Rvd^{-1}first_nt_bias_insVD        
    first_nt_bias_insDJ : ndarray
        (4,) array of the probability distribution of the indentity of the 
        first nucleotide insertion for the DJ junction.        
    zero_nt_bias_insDJ : ndarray
        (4,) array of the probability distribution of the indentity of the 
        the nucleotide BEFORE the DJ insertion. Note, as the Markov model
        at the DJ junction goes 3' to 5' this is the position AFTER the
        insertions reading left to right.
        zero_nt_bias_insVD = Rvd^{-1}first_nt_bias_insVD        
    Tvd, Svd, Dvd, lTvd, lDvd : dicts
        Dictionaries (by 'amino acid') of insertion transfer matrices 
        ((4, 4) ndarrays) for the VD junction.        
    Tdj, Sdj, Ddj, rTdj, rDdj : dicts
        Dictionaries (by 'amino acid') of insertion transfer matrices 
        ((4, 4) ndarrays) for the DJ junction.
        
    """
    
    def __init__(self, generative_model, genomic_data, alphabet_file = None):
        """Initialize PreprocessedParametersVDJ
        
        This intialization computes all of the attributes that will be needed
        for Pgen computation of CDR3 sequences generated by VDJ recombination.
        
        Parameters
        ----------
        generative_model : GenerativeModelVDJ
            VDJ generative model class containing the model parameters.            
        genomic_data : GenomicDataVDJ
            VDJ genomic data class containing the V, D, and J germline 
            sequences and info.           
        alphabet_file : str, optional
            File name (full pathing from current directory) for a custom alphabet
            definition. If no file is provided, the default alphabet is used, i.e. 
            standard amino acids, undetermined amino acids (B, J, X, and Z), and
            single codon symbols.
        
        """
        PreprocessedParameters.__init__(self, generative_model, genomic_data, alphabet_file)
        
        #Check that the generative_model and genomic_data are VDJ
        if not all([generative_model.__class__.__name__.endswith('VDJ'),genomic_data.__class__.__name__.endswith('VDJ')]):
            raise ValueError #Need VDJ model and data
        
        self.cutD_genomic_CDR3_segs = genomic_data.cutD_genomic_CDR3_segs
        self.D_allele_names = [D[0] for D in genomic_data.genD]
        
        #Set the V genomic Pi arrays
        self.PVdelV_nt_pos_vec = None
        self.PVdelV_2nd_nt_pos_per_aa_vec = None
        self.generate_PVdelV_nt_pos_vecs(generative_model, genomic_data) #Computes and sets the above attributes
        
        #Set the D genomic Pi arrays and info
        self.PD_nt_pos_vec = None
        self.PD_2nd_nt_pos_per_aa_vec = None
        self.min_delDl_given_DdelDr = None
        self.max_delDl_given_DdelDr = None
        self.zeroD_given_D = None
        self.preprocess_D_segs(generative_model, genomic_data) #Computes and sets the above attributes
        
        self.PdelDldelDr_given_D = generative_model.PdelDldelDr_given_D
        
        #Set the J genomic Pi arrays and info
        self.PJdelJ_nt_pos_vec = None
        self.PJdelJ_2nd_nt_pos_per_aa_vec = None
        self.generate_PJdelJ_nt_pos_vecs(generative_model, genomic_data) #Computes and sets the above attributes
        
        #allow for 0 prob J genes
        self.PD_given_J = np.zeros(generative_model.PDJ.shape)
        self.PD_given_J[:, np.sum(generative_model.PDJ, axis = 0) > 0] = np.multiply(generative_model.PDJ[:, np.sum(generative_model.PDJ, axis = 0) > 0], 1/np.sum(generative_model.PDJ, axis = 0)[np.sum(generative_model.PDJ, axis = 0) > 0])

        #Trim, then zeropad PinsVD and PinsDJ
        self.PinsVD = np.append(np.trim_zeros(generative_model.PinsVD, 'b'),  [0., 0., 0., 0.])
        self.PinsDJ = np.append(np.trim_zeros(generative_model.PinsDJ, 'b'),  [0., 0., 0., 0.])
        
        
        #insVD and insDJ Dinucleotide bias transition matrices
        #Note, these used to be called 'RnucleotideVD_per_nucleotideVD_5prime'
        #and 'RnucleotideDJ_per_nucleotideDJ_3prime'
        self.Rvd = generative_model.Rvd
        self.Rdj = generative_model.Rdj
        
        
        #Check if first insertion nt probability distributions are given. 
        #Note, this is not normally inferred by IGoR. If these nt bias dists
        #aren't present, use the steady-state distributions defined from Rvd
        #and Rdj. When not using the steady-state distributions we need to
        #compute the nt bias for the position previous to first_nt_bias 
        #(R^{-1}p_0)
        
        if generative_model.first_nt_bias_insVD == None:
            self.first_nt_bias_insVD = calc_steady_state_dist(self.Rvd)
            #In steady-state zero_nt_bias_insVD = first_nt_bias_insVD so no need to recalculate
            self.zero_nt_bias_insVD = self.first_nt_bias_insVD
        else:
            self.first_nt_bias_insVD = generative_model.first_nt_bias_insVD
            #Require Rvd*zero_nt_bias_insVD = first_nt_bias_insVD  --- can use pseudo-inverse
            self.zero_nt_bias_insVD = np.dot(np.linalg.pinv(self.Rvd), generative_model.first_nt_bias_insVD)
        
        if generative_model.first_nt_bias_insDJ == None:
            self.first_nt_bias_insDJ = calc_steady_state_dist(self.Rdj)
            #In steady-state zero_nt_bias_insDJ = first_nt_bias_insDJ so no need to recalculate
            self.zero_nt_bias_insDJ = self.first_nt_bias_insDJ
        else:
            self.first_nt_bias_insDJ = generative_model.first_nt_bias_insDJ
            #Require Rdj*zero_nt_bias_insDJ = first_nt_bias_insDJ  --- can use pseudo-inverse
            self.zero_nt_bias_insDJ = np.dot(np.linalg.pinv(self.Rdj), generative_model.first_nt_bias_insDJ)
            
        
        #Compute VD insertion transfer matrices
        self.Tvd = None
        self.Svd = None
        self.Dvd = None
        self.lTvd = None
        self.lDvd = None
        self.generate_VD_junction_transfer_matrices() #Computes and sets the transfer matrices
        
        #Compute DJ insertion transfer matrices
        self.Tdj = None
        self.Sdj = None
        self.Ddj = None
        self.rTdj = None
        self.rDdj = None
        self.generate_DJ_junction_transfer_matrices() #Computes and sets the transfer matrices

    def generate_PVdelV_nt_pos_vecs(self, generative_model, genomic_data):
        """Process P(V)*P(delV|V) into Pi arrays.
        
        Sets the attributes PVdelV_nt_pos_vec and PVdelV_2nd_nt_pos_per_aa_vec.
    
        Parameters
        ----------
        generative_model : GenerativeModelVDJ
            VDJ generative model class containing the model parameters.            
        genomic_data : GenomicDataVDJ
            VDJ genomic data class containing the V, D, and J germline 
            sequences and info.
        
        """
    
        cutV_genomic_CDR3_segs = genomic_data.cutV_genomic_CDR3_segs
        nt2num = {'A': 0, 'C': 1, 'G': 2, 'T': 3}
        num_del_pos = generative_model.PdelV_given_V.shape[0]
        
        PVdelV_nt_pos_vec = [[]]*len(generative_model.PV)
        PVdelV_2nd_nt_pos_per_aa_vec = [[]]*len(generative_model.PV)
        for V_in, pv in enumerate(generative_model.PV):
            current_PVdelV_nt_pos_vec = np.zeros((4, len(cutV_genomic_CDR3_segs[V_in])))
            current_PVdelV_2nd_nt_pos_per_aa_vec = {}
            for aa in self.codons_dict.keys():
                current_PVdelV_2nd_nt_pos_per_aa_vec[aa] = np.zeros((4, len(cutV_genomic_CDR3_segs[V_in])))
            for pos, nt in enumerate(cutV_genomic_CDR3_segs[V_in]):
                if len(cutV_genomic_CDR3_segs[V_in]) - pos >  num_del_pos:
                    continue
                if pos%3 == 0: #Start of a codon
                    current_PVdelV_nt_pos_vec[nt2num[nt], pos] = pv*generative_model.PdelV_given_V[len(cutV_genomic_CDR3_segs[V_in])-pos-1, V_in]    
                elif pos%3 == 1: #Mid codon position
                    for ins_nt in 'ACGT':
                        #We need to find what possible codons are allowed for any aa (or motif)
                        for aa in self.codons_dict.keys():
                            if cutV_genomic_CDR3_segs[V_in][pos-1:pos+1]+ ins_nt in self.codons_dict[aa]:
                                current_PVdelV_2nd_nt_pos_per_aa_vec[aa][nt2num[ins_nt], pos] = pv*generative_model.PdelV_given_V[len(cutV_genomic_CDR3_segs[V_in])-pos-1, V_in]            
                elif pos%3 == 2: #End of codon
                    current_PVdelV_nt_pos_vec[0, pos] = pv*generative_model.PdelV_given_V[len(cutV_genomic_CDR3_segs[V_in])-pos-1, V_in]
            PVdelV_nt_pos_vec[V_in] = current_PVdelV_nt_pos_vec
            PVdelV_2nd_nt_pos_per_aa_vec[V_in] = current_PVdelV_2nd_nt_pos_per_aa_vec
    
        self.PVdelV_nt_pos_vec = PVdelV_nt_pos_vec
        self.PVdelV_2nd_nt_pos_per_aa_vec = PVdelV_2nd_nt_pos_per_aa_vec
        
    def preprocess_D_segs(self, generative_model, genomic_data):
        """Process P(delDl, delDr|D) into Pi arrays.
        
        Sets the attributes PD_nt_pos_vec, PD_2nd_nt_pos_per_aa_vec, 
        min_delDl_given_DdelDr, max_delDl_given_DdelDr, and zeroD_given_D.
    
        Parameters
        ----------
        generative_model : GenerativeModelVDJ
            VDJ generative model class containing the model parameters.            
        genomic_data : GenomicDataVDJ
            VDJ genomic data class containing the V, D, and J germline 
            sequences and info.
        
        """
    
        cutD_genomic_CDR3_segs = genomic_data.cutD_genomic_CDR3_segs
        nt2num = {'A': 0, 'C': 1, 'G': 2, 'T': 3}
        num_dell_pos, num_delr_pos, num_D_genes = generative_model.PdelDldelDr_given_D.shape
        
        #These arrays only include the nt identity information, not the PdelDldelDr_given_D info
        PD_nt_pos_vec = [[]]*num_D_genes
        PD_2nd_nt_pos_per_aa_vec = [[]]*num_D_genes
        for D_in in range(num_D_genes):
           
            current_PD_nt_pos_vec = np.zeros((4, len(cutD_genomic_CDR3_segs[D_in])))
            current_PD_2nd_nt_pos_per_aa_vec = {}
            for aa in self.codons_dict.keys():
                current_PD_2nd_nt_pos_per_aa_vec[aa] = np.zeros((4, len(cutD_genomic_CDR3_segs[D_in])))
            
            for pos, nt in enumerate(cutD_genomic_CDR3_segs[D_in]):
                current_PD_nt_pos_vec[nt2num[nt], pos] = 1
                for ins_nt in 'ACGT':
                    for aa in self.codons_dict.keys():
                        if ins_nt + cutD_genomic_CDR3_segs[D_in][pos:pos+2] in self.codons_dict[aa]:
                            current_PD_2nd_nt_pos_per_aa_vec[aa][nt2num[ins_nt], pos] = 1
                            
            PD_nt_pos_vec[D_in] = current_PD_nt_pos_vec
            PD_2nd_nt_pos_per_aa_vec[D_in] = current_PD_2nd_nt_pos_per_aa_vec
        
        min_delDl_given_DdelDr = [[]]*num_D_genes
        max_delDl_given_DdelDr = [[]]*num_D_genes
        zeroD_given_D = [[]]*num_D_genes
        for D_in in range(num_D_genes):
            current_min_delDl_given_delDr = [0]*num_delr_pos
            current_max_delDl_given_delDr = [0]*num_delr_pos
            current_zeroD = 0
            for delr in range(num_delr_pos):
                
                if num_dell_pos > len(cutD_genomic_CDR3_segs[D_in])-delr:
                    current_zeroD += generative_model.PdelDldelDr_given_D[len(cutD_genomic_CDR3_segs[D_in])-delr, delr, D_in]
                
                dell = 0
                while generative_model.PdelDldelDr_given_D[dell, delr, D_in]==0 and dell<num_dell_pos-1:
                    dell+=1
                if generative_model.PdelDldelDr_given_D[dell, delr, D_in] == 0:
                    current_min_delDl_given_delDr[delr] = -1
                else:
                    current_min_delDl_given_delDr[delr] = dell
                if current_min_delDl_given_delDr[delr] == -1:
                    current_max_delDl_given_delDr[delr] = -1
                else:
                    dell = num_dell_pos-1
                    while generative_model.PdelDldelDr_given_D[dell, delr, D_in]==0 and dell>=0:
                        dell -= 1
                    if generative_model.PdelDldelDr_given_D[dell, delr, D_in] == 0:
                        current_max_delDl_given_delDr[delr] = -1
                    else:
                        current_max_delDl_given_delDr[delr] = dell
                
            min_delDl_given_DdelDr[D_in] = current_min_delDl_given_delDr
            max_delDl_given_DdelDr[D_in] = current_max_delDl_given_delDr
            zeroD_given_D[D_in] = current_zeroD
        
        self.PD_nt_pos_vec = PD_nt_pos_vec
        self.PD_2nd_nt_pos_per_aa_vec = PD_2nd_nt_pos_per_aa_vec
        self.min_delDl_given_DdelDr = min_delDl_given_DdelDr 
        self.max_delDl_given_DdelDr = max_delDl_given_DdelDr
        self.zeroD_given_D = zeroD_given_D
        
    def generate_PJdelJ_nt_pos_vecs(self, generative_model, genomic_data):
        """Process P(J)*P(delJ|J) into Pi arrays.
        
        Sets the attributes PJdelJ_nt_pos_vec and PJdelJ_2nd_nt_pos_per_aa_vec.
    
        Parameters
        ----------
        generative_model : GenerativeModelVDJ
            VDJ generative model class containing the model parameters.            
        genomic_data : GenomicDataVDJ
            VDJ genomic data class containing the V, D, and J germline 
            sequences and info.
        
        """
    
        cutJ_genomic_CDR3_segs = genomic_data.cutJ_genomic_CDR3_segs
        nt2num = {'A': 0, 'C': 1, 'G': 2, 'T': 3}
        num_del_pos = generative_model.PdelJ_given_J.shape[0]
    
        num_D_genes, num_J_genes = generative_model.PDJ.shape
        PJ = np.sum(generative_model.PDJ, axis = 0)
    
        PJdelJ_nt_pos_vec = [[]]*num_J_genes
        PJdelJ_2nd_nt_pos_per_aa_vec = [[]]*num_J_genes
        for J_in, pj in enumerate(PJ):
            #We include the marginal PJ here
            current_PJdelJ_nt_pos_vec = np.zeros((4, len(cutJ_genomic_CDR3_segs[J_in])))
            current_PJdelJ_2nd_nt_pos_per_aa_vec  = {}
            for aa in self.codons_dict.keys():
                current_PJdelJ_2nd_nt_pos_per_aa_vec[aa] = np.zeros((4, len(cutJ_genomic_CDR3_segs[J_in])))
    
            for pos, nt in enumerate(cutJ_genomic_CDR3_segs[J_in]):
                if pos >=  num_del_pos:
                    continue
                if (len(cutJ_genomic_CDR3_segs[J_in]) - pos)%3 == 1: #Start of a codon
                    current_PJdelJ_nt_pos_vec[nt2num[nt], pos] = pj*generative_model.PdelJ_given_J[pos, J_in]
                elif (len(cutJ_genomic_CDR3_segs[J_in]) - pos)%3 == 2: #Mid codon position
                    for ins_nt in 'ACGT':
                        #We need to find what possible codons are allowed for any aa (or motif)
                        for aa in self.codons_dict.keys():
                            if ins_nt + cutJ_genomic_CDR3_segs[J_in][pos:pos+2] in self.codons_dict[aa]:
                                current_PJdelJ_2nd_nt_pos_per_aa_vec[aa][nt2num[ins_nt], pos] = pj*generative_model.PdelJ_given_J[pos, J_in]
                                
                elif (len(cutJ_genomic_CDR3_segs[J_in]) - pos)%3 == 0: #End  of codon
                    current_PJdelJ_nt_pos_vec[0, pos] = pj*generative_model.PdelJ_given_J[pos, J_in]
            PJdelJ_nt_pos_vec[J_in] = current_PJdelJ_nt_pos_vec
            PJdelJ_2nd_nt_pos_per_aa_vec[J_in] = current_PJdelJ_2nd_nt_pos_per_aa_vec
        
        self.PJdelJ_nt_pos_vec = PJdelJ_nt_pos_vec
        self.PJdelJ_2nd_nt_pos_per_aa_vec = PJdelJ_2nd_nt_pos_per_aa_vec
        
    def generate_VD_junction_transfer_matrices(self):
        """Compute the transfer matrices for the VD junction.
        
        Sets the attributes Tvd, Svd, Dvd, lTvd, and lDvd.
        
        """  
        
        nt2num = {'A': 0, 'C': 1, 'G': 2, 'T': 3}                
        
        #Compute Tvd
        Tvd = {}
        for aa in self.codons_dict.keys():
            current_Tvd = np.zeros((4, 4))
            for init_nt in 'ACGT':
                for codon in self.codons_dict[aa]:
                    current_Tvd[nt2num[codon[2]], nt2num[init_nt]] += self.Rvd[nt2num[codon[2]],nt2num[codon[1]]]*self.Rvd[nt2num[codon[1]],nt2num[codon[0]]] * self.Rvd[nt2num[codon[0]],nt2num[init_nt]]
            Tvd[aa] = current_Tvd
            
        #Compute Svd
        Svd = {}
        for aa in self.codons_dict.keys():
            current_Svd = np.zeros((4, 4))
            for ins_nt in 'ACGT':
                if any([codon.startswith(ins_nt) for codon in self.codons_dict[aa]]):
                    current_Svd[nt2num[ins_nt], :] = self.Rvd[nt2num[ins_nt], :]
                
            Svd[aa] = current_Svd
        
        #Compute Dvd                
        Dvd = {}
        for aa in self.codons_dict.keys():
            current_Dvd = np.zeros((4, 4))
            for init_nt in 'ACGT':
                for codon in self.codons_dict[aa]:
                    current_Dvd[nt2num[codon[2]], nt2num[init_nt]] += self.Rvd[nt2num[codon[1]],nt2num[codon[0]]] * self.Rvd[nt2num[codon[0]],nt2num[init_nt]]
            Dvd[aa] = current_Dvd
     

        #Compute lTvd
        lTvd = {}
        for aa in self.codons_dict.keys():
            current_lTvd = np.zeros((4, 4))
            for codon in self.codons_dict[aa]:
                current_lTvd[nt2num[codon[2]], nt2num[codon[0]]] += self.Rvd[nt2num[codon[2]],nt2num[codon[1]]]*self.first_nt_bias_insVD[nt2num[codon[1]]]
            lTvd[aa] = current_lTvd

        
        #Compute lDvd
        lDvd = {}
        for aa in self.codons_dict.keys():
            current_lDvd = np.zeros((4, 4))
            for codon in self.codons_dict[aa]:
                current_lDvd[nt2num[codon[2]], nt2num[codon[0]]] += self.first_nt_bias_insVD[nt2num[codon[1]]]
            lDvd[aa] = current_lDvd
        
        #Set the attributes
        self.Tvd = Tvd
        self.Svd = Svd
        self.Dvd = Dvd
        self.lTvd = lTvd
        self.lDvd = lDvd
        
    def generate_DJ_junction_transfer_matrices(self):
        """Compute the transfer matrices for the VD junction.
        
        Sets the attributes Tdj, Sdj, Ddj, rTdj, and rDdj.
        
        """   
        
        nt2num = {'A': 0, 'C': 1, 'G': 2, 'T': 3}

        #Compute Tdj    
        Tdj = {}
        for aa in self.codons_dict.keys():
            current_Tdj = np.zeros((4, 4))
            for init_nt in 'ACGT':
                for codon in self.codons_dict[aa]:
                    current_Tdj[nt2num[codon[0]], nt2num[init_nt]] += self.Rdj[nt2num[codon[0]],nt2num[codon[1]]]*self.Rdj[nt2num[codon[1]],nt2num[codon[2]]] * self.Rdj[nt2num[codon[2]],nt2num[init_nt]]
            Tdj[aa] = current_Tdj
        
        #Compute Sdj
        Sdj = {}
        for aa in self.codons_dict.keys():
            current_Sdj = np.zeros((4, 4))
            for ins_nt in 'ACGT':
                if any([codon.endswith(ins_nt) for codon in self.codons_dict[aa]]):
                    current_Sdj[nt2num[ins_nt], :] = self.Rdj[nt2num[ins_nt], :]    
            Sdj[aa] = current_Sdj
        
        #Compute Ddj
        Ddj = {}
        for aa in self.codons_dict.keys():
            current_Ddj = np.zeros((4, 4))
            for init_nt in 'ACGT':
                for codon in self.codons_dict[aa]:
                    current_Ddj[nt2num[codon[0]], nt2num[init_nt]] += self.Rdj[nt2num[codon[1]],nt2num[codon[2]]] * self.Rdj[nt2num[codon[2]],nt2num[init_nt]]
            Ddj[aa] = current_Ddj
        
        #Compute rTdj
        rTdj = {}
        for aa in self.codons_dict.keys():
            current_lTdj = np.zeros((4, 4))
            for codon in self.codons_dict[aa]:
                current_lTdj[nt2num[codon[0]], nt2num[codon[2]]] += self.Rdj[nt2num[codon[0]],nt2num[codon[1]]]*self.first_nt_bias_insDJ[nt2num[codon[1]]]
            rTdj[aa] = current_lTdj
        
        #Compute rDdj
        rDdj = {}
        for aa in self.codons_dict.keys():
            current_rDdj = np.zeros((4, 4))
            for codon in self.codons_dict[aa]:
                current_rDdj[nt2num[codon[0]], nt2num[codon[2]]] += self.first_nt_bias_insDJ[nt2num[codon[1]]]
            rDdj[aa] = current_rDdj
    
        #Set the attributes
        self.Tdj = Tdj
        self.Sdj = Sdj
        self.Ddj = Ddj
        self.rTdj = rTdj
        self.rDdj = rDdj


#%% PreprocessedParametersVJ Class def
class PreprocessedParametersVJ(PreprocessedParameters):
    """Class used to preprocess the parameters of VJ recombination models.

    Attributes
    ----------
    codons_dict : dict
        Dictionary, keyed by the allowed 'amino acid' symbols with the values 
        being lists of codons corresponding to the symbol.        
    sub_codons_left : dict
        Dictionary of the 1 and 2 nucleotide prefixes (read from 5') for 
        each codon in an 'amino acid' grouping        
    sub_codons_right : dict
        Dictionary of the 1 and 2 nucleotide suffixes (read from 5') for 
        each codon in an 'amino acid' grouping
        
    V_allele_names : list of strings
        List of V allele names in genomic_data        
    d_V_usage_mask : list of int
        Default V usage mask of indices of all productive V genes/alleles.        
    V_mask_mapping : dict
        Dictionary mapping allowed keywords (strings corresponding to V gene 
        and V allele names) to the indices of the V alleles they refer to.        
    cutV_genomic_CDR3_segs : list of strings
        List of the V germline nucleotide sequences, trimmed to begin at the
        CDR3 region (includes the conserved C residue) with the maximum number
        of reverse complementary palindromic insertions appended. 
    PVJ : ndarray
        Joint probability distribution of V and J, P(V, J).
    PVdelV_nt_pos_vec : list of ndarrays
        For each V allele, format P(delV|V) into the correct form for a Pi 
        array or V(J)_{x_1}. This is only done for the first and last position 
        in each codon.    
    PVdelV_2nd_nt_pos_per_aa_vec : list of dicts
        For each V allele, and each 'amino acid', format P(V)*P(delV|V) for 
        positions in the middle of a codon into the correct form for a Pi array 
        or V(J)_{x_1} given the 'amino acid'.
        
    J_allele_names : list of strings
        List of J allele names in genomic_data        
    d_J_usage_mask : list of int
        Default J usage mask of indices of all productive J genes/alleles.        
    J_mask_mapping : dict
        Dictionary mapping allowed keywords (strings corresponding to V gene 
        and V allele names) to the indices of the V alleles they refer to.        
    cutJ_genomic_CDR3_segs : list of strings
        List of the J germline nucleotide sequences, trimmed to end at the
        CDR3 region (includes the conserved F or W residue) with the maximum 
        number of reverse complementary palindromic insertions appended.    
    PJdelJ_nt_pos_vec : list of ndarrays
        For each J allele, format P(delJ|J) into the correct form for a Pi 
        array or J^{x_2}. This is only done for the first and last position in 
        each codon.           
    PJdelJ_2nd_nt_pos_per_aa_vec : list of dicts
        For each J allele, and each 'amino acid', format P(delJ|J) for 
        positions in the middle of a codon into the correct form for a Pi array 
        or J^{x_2} given the 'amino acid'.
    
    PinsVJ : ndarray
        Probability distribution of the VJ (N) insertion sequence length        
    Rvj : ndarray
        Markov transition matrix for the VJ insertion junction.        
    first_nt_bias_insVJ : ndarray
        (4,) array of the probability distribution of the indentity of the 
        first nucleotide insertion for the VJ junction.        
    zero_nt_bias_insVJ : ndarray
        (4,) array of the probability distribution of the indentity of the 
        the nucleotide BEFORE the VJ insertion.
        zero_nt_bias_insVJ = Rvj^{-1}first_nt_bias_insVJ        
    Tvj, Svj, Dvj, lTvj, lDvj : dicts
        Dictionaries (by 'amino acid') of insertion transfer matrices 
        ((4, 4) ndarrays) for the VJ junction.
    
    """
    def __init__(self, generative_model, genomic_data, alphabet_file = None):
        """Initialize PreprocessedParametersVJ
        
        This intialization computes all of the attributes that will be needed
        for Pgen computation of CDR3 sequences generated by VJ recombination.
        
        Parameters
        ----------
        generative_model : GenerativeModelVJ
            VJ generative model class containing the model parameters.            
        genomic_data : GenomicDataVJ
            VJ genomic data class containing the V and J germline 
            sequences and info.            
        alphabet_file : str, optional
            File name (full pathing from current directory) for a custom alphabet
            definition. If no file is provided, the default alphabet is used, i.e. 
            standard amino acids, undetermined amino acids (B, J, X, and Z), and
            single codon symbols.
        
        """
        PreprocessedParameters.__init__(self, generative_model, genomic_data, alphabet_file)
        
        #Check that the generative_model and genomic_data are VDJ
        if not all([generative_model.__class__.__name__.endswith('VJ'),genomic_data.__class__.__name__.endswith('VJ')]):
            raise ValueError #Need VJ model and data
            

        self.PVJ = generative_model.PVJ
        
        #Set the V genomic Pi arrays
        self.PVdelV_nt_pos_vec = None
        self.PVdelV_2nd_nt_pos_per_aa_vec = None
        self.generate_PVdelV_nt_pos_vecs(generative_model, genomic_data) #Computes and sets the above attributes
        
        #Set the J genomic Pi arrays
        self.PJdelJ_nt_pos_vec = None
        self.PJdelJ_2nd_nt_pos_per_aa_vec = None
        self.generate_PJdelJ_nt_pos_vecs(generative_model, genomic_data) #Computes and sets the above attributes
    
        #Trim, then zeropad PinsVJ
        self.PinsVJ = np.append(np.trim_zeros(generative_model.PinsVJ, 'b'),  [0., 0., 0., 0.])
        
        
        #insVJ Dinucleotide bias transition matrix
        #Note, this used to be called 'RnucleotideVJ_per_nucleotideVJ_5prime'
        self.Rvj = generative_model.Rvj
        
        #Check if first insertion nt probability distribution is given. 
        #Note, this is not normally inferred by IGoR. If the nt bias dist isn't
        #present, use the steady-state distributions defined from Rvj.
        #When not using the steady-state distributions we need to compute the
        #nt bias for the position previous to first_nt_bias (R^{-1}p_0)
        
        if generative_model.first_nt_bias_insVJ == None:
            self.first_nt_bias_insVJ = calc_steady_state_dist(self.Rvj)
            #In steady-state zero_nt_bias_insVJ = first_nt_bias_insVJ so no need to recalculate
            self.zero_nt_bias_insVJ = self.first_nt_bias_insVJ
        else:
            self.first_nt_bias_insVJ = generative_model.first_nt_bias_insVJ
            #Require Rvj*zero_nt_bias_insVJ = first_nt_bias_insVJ  --- can use pseudo-inverse
            self.zero_nt_bias_insVJ = np.dot(np.linalg.pinv(self.Rvj), generative_model.first_nt_bias_insVJ)
            
        #Compute VJ insertion transfer matrices
        self.Tvj = None
        self.Svj = None
        self.Dvj = None
        self.lTvj = None
        self.lDvj = None
        self.generate_VJ_junction_transfer_matrices() #Computes and sets the transfer matrices
    
    def generate_PVdelV_nt_pos_vecs(self, generative_model, genomic_data):
        """Process P(delV|V) into Pi arrays.
        
        Set the attributes PVdelV_nt_pos_vec and PVdelV_2nd_nt_pos_per_aa_vec.
    
        Parameters
        ----------
        generative_model : GenerativeModelVJ
            VJ generative model class containing the model parameters.            
        genomic_data : GenomicDataVJ
            VJ genomic data class containing the V and J germline 
            sequences and info.
        
        """
    
        cutV_genomic_CDR3_segs = genomic_data.cutV_genomic_CDR3_segs
        nt2num = {'A': 0, 'C': 1, 'G': 2, 'T': 3}
        num_del_pos = generative_model.PdelV_given_V.shape[0]
        num_V_genes = generative_model.PdelV_given_V.shape[1]
        PVdelV_nt_pos_vec = [[]]*num_V_genes
        PVdelV_2nd_nt_pos_per_aa_vec = [[]]*num_V_genes
        for V_in in range(num_V_genes):
            current_PVdelV_nt_pos_vec = np.zeros((4, len(cutV_genomic_CDR3_segs[V_in])))
            current_PVdelV_2nd_nt_pos_per_aa_vec = {}
            for aa in self.codons_dict.keys():
                current_PVdelV_2nd_nt_pos_per_aa_vec[aa] = np.zeros((4, len(cutV_genomic_CDR3_segs[V_in])))
            for pos, nt in enumerate(cutV_genomic_CDR3_segs[V_in]):
                if len(cutV_genomic_CDR3_segs[V_in]) - pos >  num_del_pos:
                    continue
                if pos%3 == 0: #Start of a codon
                    current_PVdelV_nt_pos_vec[nt2num[nt], pos] = generative_model.PdelV_given_V[len(cutV_genomic_CDR3_segs[V_in])-pos-1, V_in]    
                elif pos%3 == 1: #Mid codon position
                    for ins_nt in 'ACGT':
                        #We need to find what possible codons are allowed for any aa (or motif)
                        for aa in self.codons_dict.keys():
                            if cutV_genomic_CDR3_segs[V_in][pos-1:pos+1]+ ins_nt in self.codons_dict[aa]:
                                current_PVdelV_2nd_nt_pos_per_aa_vec[aa][nt2num[ins_nt], pos] = generative_model.PdelV_given_V[len(cutV_genomic_CDR3_segs[V_in])-pos-1, V_in]            
                elif pos%3 == 2: #End of codon
                    current_PVdelV_nt_pos_vec[0, pos] = generative_model.PdelV_given_V[len(cutV_genomic_CDR3_segs[V_in])-pos-1, V_in]
            PVdelV_nt_pos_vec[V_in] = current_PVdelV_nt_pos_vec
            PVdelV_2nd_nt_pos_per_aa_vec[V_in] = current_PVdelV_2nd_nt_pos_per_aa_vec
    
        
        self.PVdelV_nt_pos_vec = PVdelV_nt_pos_vec
        self.PVdelV_2nd_nt_pos_per_aa_vec = PVdelV_2nd_nt_pos_per_aa_vec

    def generate_PJdelJ_nt_pos_vecs(self, generative_model, genomic_data):
        """Process P(delJ|J) into Pi arrays.
        
        Set the attributes PJdelJ_nt_pos_vec and PJdelJ_2nd_nt_pos_per_aa_vec.
    
        Parameters
        ----------
        generative_model : GenerativeModelVJ
            VJ generative model class containing the model parameters.            
        genomic_data : GenomicDataVJ
            VJ genomic data class containing the V and J germline 
            sequences and info.
        
        """
        
        cutJ_genomic_CDR3_segs = genomic_data.cutJ_genomic_CDR3_segs
        nt2num = {'A': 0, 'C': 1, 'G': 2, 'T': 3}
        num_del_pos = generative_model.PdelJ_given_J.shape[0]
    
        num_D_genes, num_J_genes = generative_model.PVJ.shape
    
        PJdelJ_nt_pos_vec = [[]]*num_J_genes
        PJdelJ_2nd_nt_pos_per_aa_vec = [[]]*num_J_genes
        for J_in in range(num_J_genes):
            current_PJdelJ_nt_pos_vec = np.zeros((4, len(cutJ_genomic_CDR3_segs[J_in])))
            current_PJdelJ_2nd_nt_pos_per_aa_vec  = {}
            for aa in self.codons_dict.keys():
                current_PJdelJ_2nd_nt_pos_per_aa_vec[aa] = np.zeros((4, len(cutJ_genomic_CDR3_segs[J_in])))
    
            for pos, nt in enumerate(cutJ_genomic_CDR3_segs[J_in]):
                if pos >=  num_del_pos:
                    continue
                if (len(cutJ_genomic_CDR3_segs[J_in]) - pos)%3 == 1: #Start of a codon
                    current_PJdelJ_nt_pos_vec[nt2num[nt], pos] = generative_model.PdelJ_given_J[pos, J_in]
                elif (len(cutJ_genomic_CDR3_segs[J_in]) - pos)%3 == 2: #Mid codon position
                    for ins_nt in 'ACGT':
                        #We need to find what possible codons are allowed for any aa (or motif)
                        for aa in self.codons_dict.keys():
                            if ins_nt + cutJ_genomic_CDR3_segs[J_in][pos:pos+2] in self.codons_dict[aa]:
                                current_PJdelJ_2nd_nt_pos_per_aa_vec[aa][nt2num[ins_nt], pos] = generative_model.PdelJ_given_J[pos, J_in]
                                
                elif (len(cutJ_genomic_CDR3_segs[J_in]) - pos)%3 == 0: #End  of codon
                    current_PJdelJ_nt_pos_vec[0, pos] = generative_model.PdelJ_given_J[pos, J_in]
            PJdelJ_nt_pos_vec[J_in] = current_PJdelJ_nt_pos_vec
            PJdelJ_2nd_nt_pos_per_aa_vec[J_in] = current_PJdelJ_2nd_nt_pos_per_aa_vec
            
        self.PJdelJ_nt_pos_vec = PJdelJ_nt_pos_vec
        self.PJdelJ_2nd_nt_pos_per_aa_vec = PJdelJ_2nd_nt_pos_per_aa_vec
        
        
    def generate_VJ_junction_transfer_matrices(self):
        """Compute the transfer matrices for the VJ junction.
        
        Sets the attributes Tvj, Svj, Dvj, lTvj, and lDvj.
        
        """    
        
        nt2num = {'A': 0, 'C': 1, 'G': 2, 'T': 3}                
        
        #Compute Tvj
        Tvj = {}
        for aa in self.codons_dict.keys():
            current_Tvj = np.zeros((4, 4))
            for init_nt in 'ACGT':
                for codon in self.codons_dict[aa]:
                    current_Tvj[nt2num[codon[2]], nt2num[init_nt]] += self.Rvj[nt2num[codon[2]],nt2num[codon[1]]]*self.Rvj[nt2num[codon[1]],nt2num[codon[0]]] * self.Rvj[nt2num[codon[0]],nt2num[init_nt]]
            Tvj[aa] = current_Tvj
    
        #Compute Svj
        Svj = {}
        for aa in self.codons_dict.keys():
            current_Svj = np.zeros((4, 4))
            for ins_nt in 'ACGT':
                if any([codon.startswith(ins_nt) for codon in self.codons_dict[aa]]):
                    current_Svj[nt2num[ins_nt], :] = self.Rvj[nt2num[ins_nt], :]             
            Svj[aa] = current_Svj
        
        #Compute Dvj               
        Dvj = {}    
        for aa in self.codons_dict.keys():
            current_Dvj = np.zeros((4, 4))
            for init_nt in 'ACGT':
                for codon in self.codons_dict[aa]:
                    current_Dvj[nt2num[codon[2]], nt2num[init_nt]] += self.Rvj[nt2num[codon[1]],nt2num[codon[0]]] * self.Rvj[nt2num[codon[0]],nt2num[init_nt]]
            Dvj[aa] = current_Dvj

        #Compute lTvj
        lTvj = {}
        for aa in self.codons_dict.keys():
            current_lTvj = np.zeros((4, 4))
            for codon in self.codons_dict[aa]:
                current_lTvj[nt2num[codon[2]], nt2num[codon[0]]] += self.Rvj[nt2num[codon[2]],nt2num[codon[1]]]*self.first_nt_bias_insVJ[nt2num[codon[1]]]
            lTvj[aa] = current_lTvj

        #Compute lDvj        
        lDvj = {}    
        for aa in self.codons_dict.keys():
            current_lDvj = np.zeros((4, 4))
            for codon in self.codons_dict[aa]:
                current_lDvj[nt2num[codon[2]], nt2num[codon[0]]] += self.first_nt_bias_insVJ[nt2num[codon[1]]]
            lDvj[aa] = current_lDvj
    

        #Set the attributes
        self.Tvj = Tvj
        self.Svj = Svj
        self.Dvj = Dvj
        self.lTvj = lTvj
        self.lDvj = lDvj